from yowsup.structs import ProtocolEntity, ProtocolTreeNode
import sys
class EncProtocolEntity(ProtocolEntity):
    TYPE_PKMSG  = "pkmsg"
    TYPE_MSG    = "msg"
    TYPE_SKMSG  = "skmsg"
    TYPES = (TYPE_PKMSG, TYPE_MSG, TYPE_SKMSG)

    def __init__(self, type, version, data, mediaType = None, jid = None):
        assert type in self.__class__.TYPES, "Unknown message enc type %s" % type
        super(EncProtocolEntity, self).__init__("enc")
        self.type = type
        self.version = int(version)
        self.data = data
        self.mediaType = mediaType
        self.jid = jid

    def getType(self):
        return self.type

    def getVersion(self):
        return self.version

    def getData(self):
        return self.data

    def getMediaType(self):
        return self.mediaType

    def getJid(self):
        return self.jid

    def toProtocolTreeNode(self):
        attribs = {"type": self.type, "v": str(self.version)}
        if self.mediaType:
            attribs["mediatype"] = self.mediaType
        encNode =  ProtocolTreeNode("enc", attribs, data = self.data)
        if self.jid:
            return ProtocolTreeNode("to", {"jid": self.jid}, [encNode])
        return encNode

    @staticmethod
    def fromProtocolTreeNode(node):
        return EncProtocolEntity(node["type"], node["v"], node.data.encode('latin-1') if sys.version_info >= (3,0) else node.data, node["mediatype"])



from yowsup.layers.protocol_media.protocolentities.attributes.context_info import ContextInfo


class MediaAttributes(object):
    def __init__(self, context_info=None):
        """
        :type context_info: ContextInfo | None
        """
        if context_info:
            assert type(context_info) is ContextInfo
        self._context_info = context_info  # type: ContextInfo | None

    @property
    def context_info(self):
        return self._context_info



from .message import MessageProtocolEntity
from .proto import ProtoProtocolEntity
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message
import logging

logger = logging.getLogger(__name__)


class ProtomessageProtocolEntity(MessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}"
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
            <proto>
                {{SERIALIZE_PROTO_DATA}}
            </proto>
    </message>
    '''
    def __init__(self, messageType, messageAttributes):
        super(ProtomessageProtocolEntity, self).__init__(messageType, messageAttributes)
        self._proto = Message()

    def __str__(self):
        out = super(ProtomessageProtocolEntity, self).__str__()
        return "%s\nproto=%s" % (out, self._proto)

    @property
    def proto(self):
        return self._proto

    def deserializeProtoData(self, protoData):
        m = Message()
        m.ParseFromString(protoData)
        return m

    def toProtocolTreeNode(self):
        node = super(ProtomessageProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtoProtocolEntity(self._proto.SerializeToString()).toProtocolTreeNode())

        return node

    def setProtoFromData(self, protoData):
        self._proto = self.deserializeProtoData(protoData)

    @classmethod
    def fromProtocolTreeNode(cls, node):
        entity = MessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__= cls
        entity.setProtoFromData(node.getChild("proto").getData())
        return entity


from yowsup.config.base.transform import ConfigTransform


class MapTransform(ConfigTransform):

    def __init__(self, transform_map=None, reverse_map=None):
        """
        :param transform_map:
        :type transform_map: function | None
        :param reverse_map:
        :type reverse_map: function | None
        """
        self._transform_map = transform_map  # type: function | None
        self._reverse_map = reverse_map  # type: function | None

    def transform(self, data):
        if self._transform_map is not None:
            out = {}
            for key, val in data.items():
                key, val = self._transform_map(key, val)
                out[key] = val
            return out
        return data

    def reverse(self, data):
        if self._reverse_map is not None:
            out = {}
            for key, val in data.items():
                key, val = self._reverse_map(key, val)
                out[key] = val
            return out
        return data



import os


class DocumentAttributes(object):
    def __init__(self, file_name, file_length, title=None, page_count=None, jpeg_thumbnail=None):
        self._file_name = file_name
        self._file_length = file_length
        self._title = title
        self._page_count = page_count
        self._jpeg_thumbnail = jpeg_thumbnail

    @property
    def file_name(self):
        return self._file_name

    @property
    def file_length(self):
        return self._file_length

    @property
    def title(self):
        return self._title

    @property
    def page_count(self):
        return self._page_count

    @property
    def jpeg_thumbnail(self):
        return self._jpeg_thumbnail

    @staticmethod
    def from_filepath(filepath):
        return DocumentAttributes(
            os.path.basename(filepath),
            os.path.getsize(filepath)
        )



from yowsup.config.v1.config import Config
from yowsup.config.transforms.dict_keyval import DictKeyValTransform
from yowsup.config.transforms.dict_json import DictJsonTransform
from yowsup.config.v1.serialize import ConfigSerialize
from yowsup.common.tools import StorageTools
import logging
import os

logger = logging.getLogger(__name__)


class ConfigManager(object):
    NAME_FILE_CONFIG = "config"

    TYPE_KEYVAL = 1
    TYPE_JSON = 2

    TYPE_NAMES = {
        TYPE_KEYVAL: "keyval",
        TYPE_JSON: "json"
    }

    MAP_EXT = {
        "yo": TYPE_KEYVAL,
        "json": TYPE_JSON,
    }

    TYPES = {
        TYPE_KEYVAL: DictKeyValTransform,
        TYPE_JSON: DictJsonTransform
    }

    def load(self, username):
        """
        :param username:
        :type username:
        :return:
        :rtype:
        """
        config_dir = StorageTools.getStorageForPhone(username)
        logger.debug("Detecting config for username=%s, dir=%s" % (username, config_dir))
        exhausted = []
        for ftype in self.MAP_EXT:
            if len(ftype):
                fname = (self.NAME_FILE_CONFIG + "." + ftype)
            else:
                fname = self.NAME_FILE_CONFIG

            fpath = os.path.join(config_dir, fname)
            logger.debug("Trying %s" % fpath)
            if os.path.isfile(fpath):
                return self.load_path(fpath)

            exhausted.append(fpath)

        logger.error("Could not find a config for username=%s, paths checked: %s" % (username, ":".join(exhausted)))

    def _type_to_str(self, type):
        """
        :param type:
        :type type: int
        :return:
        :rtype:
        """
        for key, val in self.TYPE_NAMES.items():
            if key == type:
                return val

    def load_path(self, path):
        """
        :param path:
        :type path:
        :return:
        :rtype:
        """
        logger.debug("load_path(path=%s)" % path)
        if os.path.isfile(path):
            configtype = self.guess_type(path)
            logger.debug("Detected config type: %s" % self._type_to_str(configtype))
            if configtype in self.TYPES:
                logger.debug("Opening config for reading")
                with open(path, 'r') as f:
                    data = f.read()
                datadict = self.TYPES[configtype]().reverse(data)
                return self.load_data(datadict)
            else:
                raise ValueError("Unsupported config type")
        else:
            logger.warn("load_path couldn't find the path: %s" % path)

    def load_data(self, datadict):
        logger.debug("Loading config")
        return ConfigSerialize(Config).deserialize(datadict)

    def guess_type(self, config_path):
        dissected = os.path.splitext(config_path)
        if len(dissected) > 1:
            ext = dissected[1][1:].lower()
            config_type = self.MAP_EXT[ext] if ext in self.MAP_EXT else None
        else:
            config_type = None

        if config_type is not None:
            return config_type
        else:
            logger.debug("Trying auto detect config type by parsing")
            with open(config_path, 'r') as f:
                data = f.read()
            for config_type, transform in self.TYPES.items():
                config_type_str = self.TYPE_NAMES[config_type]
                try:
                    logger.debug("Trying to parse as %s" % config_type_str)
                    if transform().reverse(data):
                        logger.debug("Successfully detected %s as config type for %s" % (config_type_str, config_path))
                        return config_type
                except Exception as ex:
                    logger.debug("%s was not parseable as %s, reason: %s" % (config_path, config_type_str, ex))

    def get_str_transform(self, serialize_type):
        if serialize_type in self.TYPES:
            return self.TYPES[serialize_type]()

    def config_to_str(self, config, serialize_type=TYPE_JSON):
        transform = self.get_str_transform(serialize_type)
        if transform is not None:
            return transform.transform(ConfigSerialize(config.__class__).serialize(config))

        raise ValueError("unrecognized serialize_type=%d" % serialize_type)

    def save(self, config, serialize_type=TYPE_JSON, dest=None):
        outputdata = self.config_to_str(config, serialize_type)
        if dest is None:
            StorageTools.writePhoneConfig(config.phone, outputdata)
        else:
            with open(dest, 'wb') as outputfile:
                outputfile.write(outputdata)



class TokenDictionary:
    FLAG_SEGMENTED = 0x1
    FLAG_DEFLATE = 0x2

    def __init__(self):
        self.dictionary = [
          '',
          '',
          '',
          'account',
          'ack',
          'action',
          'active',
          'add',
          'after',
          'all',
          'allow',
          'apple',
          'audio',
          'auth',
          'author',
          'available',
          'bad-protocol',
          'bad-request',
          'before',
          'bits',
          'body',
          'broadcast',
          'cancel',
          'category',
          'challenge',
          'chat',
          'clean',
          'code',
          'composing',
          'config',
          'contacts',
          'count',
          'create',
          'creation',
          'debug',
          'default',
          'delete',
          'delivery',
          'delta',
          'deny',
          'digest',
          'dirty',
          'duplicate',
          'elapsed',
          'enable',
          'encoding',
          'encrypt',
          'error',
          'event',
          'expiration',
          'expired',
          'fail',
          'failure',
          'false',
          'favorites',
          'feature',
          'features',
          'feature-not-implemented',
          'field',
          'file',
          'filehash',
          'first',
          'free',
          'from',
          'g.us',
          'gcm',
          'get',
          'google',
          'group',
          'groups',
          'groups_v2',
          'http://etherx.jabber.org/streams',
          'http://jabber.org/protocol/chatstates',
          'ib',
          'id',
          'image',
          'img',
          'index',
          'internal-server-error',
          'ip',
          'iq',
          'item-not-found',
          'item',
          'jabber:iq:last',
          'jabber:iq:privacy',
          'jabber:x:event',
          'jid',
          'kind',
          'last',
          'leave',
          'list',
          'max',
          'mechanism',
          'media',
          'message_acks',
          'message',
          'method',
          'microsoft',
          'mimetype',
          'missing',
          'modify',
          'msg',
          'mute',
          'name',
          'nokia',
          'none',
          'not-acceptable',
          'not-allowed',
          'not-authorized',
          'notification',
          'notify',
          'off',
          'offline',
          'order',
          'owner',
          'owning',
          'p_o',
          'p_t',
          'paid',
          'participant',
          'participants',
          'participating',
          'paused',
          'picture',
          'pin',
          'ping',
          'pkmsg',
          'platform',
          'port',
          'presence',
          'preview',
          'probe',
          'prop',
          'props',
          'qcount',
          'query',
          'raw',
          'read',
          'readreceipts',
          'reason',
          'receipt',
          'relay',
          'remote-server-timeout',
          'remove',
          'request',
          'required',
          'resource-constraint',
          'resource',
          'response',
          'result',
          'retry',
          'rim',
          's_o',
          's_t',
          's.us',
          's.whatsapp.net',
          'seconds',
          'server-error',
          'server',
          'service-unavailable',
          'set',
          'show',
          'silent',
          'size',
          'skmsg',
          'stat',
          'state',
          'status',
          'stream:error',
          'stream:features',
          'subject',
          'subscribe',
          'success',
          'sync',
          't',
          'text',
          'timeout',
          'timestamp',
          'tizen',
          'to',
          'true',
          'type',
          'unavailable',
          'unsubscribe',
          'upgrade',
          'uri',
          'url',
          'urn:ietf:params:xml:ns:xmpp-sasl',
          'urn:ietf:params:xml:ns:xmpp-stanzas',
          'urn:ietf:params:xml:ns:xmpp-streams',
          'urn:xmpp:ping',
          'urn:xmpp:whatsapp:account',
          'urn:xmpp:whatsapp:dirty',
          'urn:xmpp:whatsapp:mms',
          'urn:xmpp:whatsapp:push',
          'urn:xmpp:whatsapp',
          'user',
          'user-not-found',
          'v',
          'value',
          'version',
          'voip',
          'w:g',
          'w:p:r',
          'w:p',
          'w:profile:picture',
          'w',
          'wait',
          'WAUTH-2',
          'xmlns:stream',
          'xmlns',
          '1',
          'chatstate',
          'crypto',
          'phash',
          'enc',
          'class',
          'off_cnt',
          'w:g2',
          'promote',
          'demote',
          'creator',
          'background',
          'backoff',
          'chunked',
          'context',
          'full',
          'in',
          'interactive',
          'out',
          'registration',
          'sid',
          'urn:xmpp:whatsapp:sync',
          'flt',
          's16',
          'u8',
        ]

        self.secondaryDictionary = [
          'adpcm',
          'amrnb',
          'amrwb',
          'mp3',
          'pcm',
          'qcelp',
          'wma',
          'h263',
          'h264',
          'jpeg',
          'mpeg4',
          'wmv',
          'audio/3gpp',
          'audio/aac',
          'audio/amr',
          'audio/mp4',
          'audio/mpeg',
          'audio/ogg',
          'audio/qcelp',
          'audio/wav',
          'audio/webm',
          'audio/x-caf',
          'audio/x-ms-wma',
          'image/gif',
          'image/jpeg',
          'image/png',
          'video/3gpp',
          'video/avi',
          'video/mp4',
          'video/mpeg',
          'video/quicktime',
          'video/x-flv',
          'video/x-ms-asf',
          '302',
          '400',
          '401',
          '402',
          '403',
          '404',
          '405',
          '406',
          '407',
          '409',
          '410',
          '500',
          '501',
          '503',
          '504',
          'abitrate',
          'acodec',
          'app_uptime',
          'asampfmt',
          'asampfreq',
          'clear',
          'conflict',
          'conn_no_nna',
          'cost',
          'currency',
          'duration',
          'extend',
          'fps',
          'g_notify',
          'g_sound',
          'gone',
          'google_play',
          'hash',
          'height',
          'invalid',
          'jid-malformed',
          'latitude',
          'lc',
          'lg',
          'live',
          'location',
          'log',
          'longitude',
          'max_groups',
          'max_participants',
          'max_subject',
          'mode',
          'napi_version',
          'normalize',
          'orighash',
          'origin',
          'passive',
          'password',
          'played',
          'policy-violation',
          'pop_mean_time',
          'pop_plus_minus',
          'price',
          'pricing',
          'redeem',
          'Replaced by new connection',
          'resume',
          'signature',
          'sound',
          'source',
          'system-shutdown',
          'username',
          'vbitrate',
          'vcard',
          'vcodec',
          'video',
          'width',
          'xml-not-well-formed',
          'checkmarks',
          'image_max_edge',
          'image_max_kbytes',
          'image_quality',
          'ka',
          'ka_grow',
          'ka_shrink',
          'newmedia',
          'library',
          'caption',
          'forward',
          'c0',
          'c1',
          'c2',
          'c3',
          'clock_skew',
          'cts',
          'k0',
          'k1',
          'login_rtt',
          'm_id',
          'nna_msg_rtt',
          'nna_no_off_count',
          'nna_offline_ratio',
          'nna_push_rtt',
          'no_nna_con_count',
          'off_msg_rtt',
          'on_msg_rtt',
          'stat_name',
          'sts',
          'suspect_conn',
          'lists',
          'self',
          'qr',
          'web',
          'w:b',
          'recipient',
          'w:stats',
          'forbidden',
          'max_list_recipients',
          'en-AU',
          'en-GB',
          'es-MX',
          'pt-PT',
          'zh-Hans',
          'zh-Hant',
          'relayelection',
          'relaylatency',
          'interruption',
          'Bell.caf',
          'Boing.caf',
          'Glass.caf',
          'Harp.caf',
          'TimePassing.caf',
          'Tri-tone.caf',
          'Xylophone.caf',
          'aurora.m4r',
          'bamboo.m4r',
          'chord.m4r',
          'circles.m4r',
          'complete.m4r',
          'hello.m4r',
          'input.m4r',
          'keys.m4r',
          'note.m4r',
          'popcorn.m4r',
          'pulse.m4r',
          'synth.m4r',
          'Apex.m4r',
          'Beacon.m4r',
          'Bulletin.m4r',
          'By The Seaside.m4r',
          'Chimes.m4r',
          'Circuit.m4r',
          'Constellation.m4r',
          'Cosmic.m4r',
          'Crystals.m4r',
          'Hillside.m4r',
          'Illuminate.m4r',
          'Night Owl.m4r',
          'Opening.m4r',
          'Playtime.m4r',
          'Presto.m4r',
          'Radar.m4r',
          'Radiate.m4r',
          'Ripples.m4r',
          'Sencha.m4r',
          'Signal.m4r',
          'Silk.m4r',
          'Slow Rise.m4r',
          'Stargaze.m4r',
          'Summit.m4r',
          'Twinkle.m4r',
          'Uplift.m4r',
          'Waves.m4r',
          'eligible',
          'planned',
          'current',
          'future',
          'disable',
          'expire',
          'start',
          'stop',
          'accuracy',
          'speed',
          'bearing',
          'recording',
          'key',
          'identity',
          'w:gp2',
          'admin',
          'locked',
          'unlocked',
          'new',
          'battery',
          'archive',
          'adm',
          'plaintext_size',
          'plaintext_disabled',
          'plaintext_reenable_threshold',
          'compressed_size',
          'delivered',
          'everyone',
          'transport',
          'mspes',
          'e2e_groups',
          'e2e_images',
          'encr_media',
          'encrypt_v2',
          'encrypt_image',
          'encrypt_sends_push',
          'force_long_connect',
          'audio_opus',
          'video_max_edge',
          'call-id',
          'call',
          'preaccept',
          'accept',
          'offer',
          'reject',
          'busy',
          'te',
          'terminate',
          'begin',
          'end',
          'opus',
          'rtt',
          'token',
          'priority',
          'p2p',
          'rate',
          'amr',
          'ptt',
          'srtp',
          'os',
          'browser',
          'encrypt_group_gen2',
          "encrypt_audio",
          "encrypt_blist",
          "encrypt_contact",
          "encrypt_location",
          "encrypt_url",
          "encrypt_video",
          "doc_types",
          "upload_oom_hprof_enabled",
          "tos",
          "client",
          "e2e_audio",
          "e2e_blists",
          "e2e_video",
          "document",
          "contact",
          "file_max_size",
          "small_call_btn",
          "enable_mp4_operations_mux",
          "android_vacuuming_enabled",
          "android_vacuum_experiment_enabled",
          "aec",
          "agc",
          "options",
          "encode",
          "bwe",
          "rc",
          "ns",
          "ec_threshold",
          "ec_off_threshold",
          "algorithm",
          "targetlevel",
          "compressiongain",
          "limiterenable",
          "cbr",
          "complexity",
          "minfpp",
          "maxrtt",
          "low_data_usage_bitrate",
          "usync",
          "refresh",
          "media_max_autodownload"
        ]


    def getToken(self, index, secondary = False):
        targetDict = self.dictionary
        if secondary:
            targetDict = self.secondaryDictionary
        elif index > 236 and index < (236 + len(self.secondaryDictionary)):
            targetDict = self.secondaryDictionary
            index = index - 237
        if index < 0 or index > len(targetDict) - 1:
            return None

        return targetDict[index]

    def getIndex(self, token):
        if token in self.dictionary:
            return (self.dictionary.index(token), False)
        elif token in self.secondaryDictionary:
            return (self.secondaryDictionary.index(token), True)

        return None



from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_downloadablemedia \
    import DownloadableMediaMessageAttributes
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_audio \
    import AudioAttributes
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes


class AudioDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    def __init__(self, audio_attrs, downloadablemedia_attrs, message_attrs):
        """
        :type audio_attrs: AudioAttributes
        :type downloadablemedia_attrs: DownloadableMediaMessageAttributes
        :type message_attrs: MessageAttributes
        """
        super(AudioDownloadableMediaMessageProtocolEntity, self).__init__(
            "audio", downloadablemedia_attrs, message_attrs
        )
        self.seconds = audio_attrs.seconds
        self.ptt = audio_attrs.ptt

    @property
    def proto(self):
        return self._proto.audio_message

    @property
    def seconds(self):
        return self.proto.seconds

    @seconds.setter
    def seconds(self, value):
        """
        :type value: int
        """
        self.proto.seconds = value

    @property
    def ptt(self):
        return self.proto.ptt

    @ptt.setter
    def ptt(self, value):
        """
        :type value: bool
        """
        self.proto.ptt = value



from yowsup.common import YowConstants
from yowsup.layers.protocol_notifications.protocolentities import NotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode


class IdentityChangeEncryptNotification(NotificationProtocolEntity):
    """
    <notification t="1419824928" id="2451228097" from="s.whatsapp.net" type="encrypt">
        <identity />
    </notification>
    """
    def __init__(self, timestamp, _id = None, notify = None, offline = None):
        super(IdentityChangeEncryptNotification, self).__init__(
            "encrypt", _id, YowConstants.WHATSAPP_SERVER, timestamp, notify, offline
        )

    def toProtocolTreeNode(self):
        node = super(IdentityChangeEncryptNotification, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("identity"))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = NotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = IdentityChangeEncryptNotification
        return entity



from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes
from .message_media import MediaMessageProtocolEntity
from .attributes.attributes_location import LocationAttributes
from .attributes.attributes_media import MediaAttributes


class LocationMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    def __init__(self, location_attrs, media_message_attrs, message_attrs):
        # type: (LocationAttributes, MediaAttributes, MessageAttributes) -> None
        super(LocationMediaMessageProtocolEntity, self).__init__("location", media_message_attrs, message_attrs)
        self.degrees_latitude = location_attrs.degrees_latitude
        self.degrees_longitude = location_attrs.degrees_longitude
        self.name = location_attrs.name
        self.address = location_attrs.address
        self.url = location_attrs.url
        self.duration = location_attrs.duration
        self.accuracy_in_meters = location_attrs.accuracy_in_meters
        self.speed_in_mps = location_attrs.speed_in_mps
        self.degrees_clockwise_from_magnetic_north = location_attrs.degrees_clockwise_from_magnetic_north
        self.axolotl_sender_key_distribution_message = location_attrs.axolotl_sender_key_distribution_message
        self.jpeg_thumbnail = location_attrs.jpeg_thumbnail

    @property
    def proto(self):
        return self._proto.location_message

    @property
    def degrees_latitude(self):
        return self.proto.degrees_latitude

    @degrees_latitude.setter
    def degrees_latitude(self, value):
        self.proto.degrees_latitude = value

    @property
    def degrees_longitude(self):
        return self.proto.degrees_longitude

    @degrees_longitude.setter
    def degrees_longitude(self, value):
        self.proto.degrees_longitude = value

    @property
    def name(self):
        return self.proto.name

    @name.setter
    def name(self, value):
        self.proto.name = value

    @property
    def address(self):
        return self.proto.addrees

    @address.setter
    def address(self, value):
        self.proto.address = value

    @property
    def url(self):
        return self.proto.url

    @url.setter
    def url(self, value):
        self.proto.url = value

    @property
    def duration(self):
        return self.proto.duration

    @duration.setter
    def duration(self, value):
        self.proto.duration = value

    @property
    def accuracy_in_meters(self):
        return self.proto.accuracy_in_meters

    @accuracy_in_meters.setter
    def accuracy_in_meters(self, value):
        self.proto.accuracy_in_meters = value

    @property
    def speed_in_mps(self):
        return self.proto.speed_in_mps

    @speed_in_mps.setter
    def speed_in_mps(self, value):
        self.proto.speed_in_mps = value

    @property
    def degrees_clockwise_from_magnetic_north(self):
        return self.proto.degrees_clockwise_from_magnetic_north

    @degrees_clockwise_from_magnetic_north.setter
    def degrees_clockwise_from_magnetic_north(self, value):
        self.proto.degrees_clockwise_from_magnetic_north = value

    @property
    def axolotl_sender_key_distribution_message(self):
        return self.proto.axolotl_sender_key_distribution_message

    @axolotl_sender_key_distribution_message.setter
    def axolotl_sender_key_distribution_message(self, value):
        self.proto.axolotl_sender_key_distribution_message = value

    @property
    def jpeg_thumbnail(self):
        return self.proto.jpeg_thumbnail

    @jpeg_thumbnail.setter
    def jpeg_thumbnail(self, value):
        self.proto.jpeg_thumbnail = value



import os
from .constants import YowConstants
import codecs, sys
import logging
import tempfile
import base64
import hashlib
import os.path, mimetypes
import uuid
from consonance.structs.keypair import KeyPair
from appdirs import user_config_dir

from .optionalmodules import PILOptionalModule, FFVideoOptionalModule

logger = logging.getLogger(__name__)

class Jid:
    @staticmethod
    def normalize(number):
        if '@' in number:
            return number
        elif "-" in number:
            return "%s@%s" % (number, YowConstants.WHATSAPP_GROUP_SERVER)
        return "%s@%s" % (number, YowConstants.WHATSAPP_SERVER)

class HexTools:
    decode_hex = codecs.getdecoder("hex_codec")
    @staticmethod
    def decodeHex(hexString):
        result = HexTools.decode_hex(hexString)[0]
        if sys.version_info >= (3,0):
            result = result.decode('latin-1')
        return result

class WATools:
    @staticmethod
    def generateIdentity():
        return os.urandom(20)

    @classmethod
    def generatePhoneId(cls):
        """
        :return:
        :rtype: str
        """
        return str(cls.generateUUID())

    @classmethod
    def generateDeviceId(cls):
        """
        :return:
        :rtype: bytes
        """
        return cls.generateUUID().bytes

    @classmethod
    def generateUUID(cls):
        """
        :return:
        :rtype: uuid.UUID
        """
        return uuid.uuid4()

    @classmethod
    def generateKeyPair(cls):
        """
        :return:
        :rtype: KeyPair
        """
        return KeyPair.generate()

    @staticmethod
    def getFileHashForUpload(filePath):
        sha1 = hashlib.sha256()
        f = open(filePath, 'rb')
        try:
            sha1.update(f.read())
        finally:
            f.close()
        b64Hash = base64.b64encode(sha1.digest())
        return b64Hash if type(b64Hash) is str else b64Hash.decode()


class StorageTools:
    NAME_CONFIG = "config.json"

    @staticmethod
    def constructPath(*path):
        path = os.path.join(*path)
        fullPath = os.path.join(user_config_dir(YowConstants.YOWSUP), path)
        if not os.path.exists(os.path.dirname(fullPath)):
            os.makedirs(os.path.dirname(fullPath))
        return fullPath

    @staticmethod
    def getStorageForPhone(phone):
        if type(phone) is not str:
            phone = str(phone)
        return StorageTools.constructPath(phone + '/')

    @staticmethod
    def writePhoneData(phone, name, val):
        logger.debug("writePhoneData(phone=%s, name=%s, val=[omitted])" % (phone, name))
        path = os.path.join(StorageTools.getStorageForPhone(phone), name)
        logger.debug("Writing %s" % path)

        with open(path, 'w' if type(val) is str else 'wb') as attrFile:
            attrFile.write(val)

    @staticmethod
    def readPhoneData(phone, name, default=None):
        logger.debug("readPhoneData(phone=%s, name=%s)" % (phone, name))
        path = StorageTools.getStorageForPhone(phone)
        dataFilePath = os.path.join(path, name)
        if os.path.isfile(dataFilePath):
            logger.debug("Reading %s" % dataFilePath)
            with open(dataFilePath, 'rb') as attrFile:
                return attrFile.read()
        else:
            logger.debug("%s does not exist" % dataFilePath)

        return default

    @classmethod
    def writeIdentity(cls, phone, identity):
        cls.writePhoneData(phone, 'id', identity)

    @classmethod
    def getIdentity(cls, phone):
        return cls.readPhoneData(phone, 'id')

    @classmethod
    def writePhoneConfig(cls, phone, config):
        cls.writePhoneData(phone, cls.NAME_CONFIG, config)

    @classmethod
    def readPhoneConfig(cls, phone, config):
        return cls.readPhoneData(phone, cls.NAME_CONFIG)


class ImageTools:
    @staticmethod
    def scaleImage(infile, outfile, imageFormat, width, height):
        with PILOptionalModule() as imp:
            Image = imp("Image")
            im = Image.open(infile)
            #Convert P mode images
            if im.mode != "RGB":
                im = im.convert("RGB")
            im.thumbnail((width, height))
            im.save(outfile, imageFormat)
            return True
        return False

    @staticmethod
    def getImageDimensions(imageFile):
        with PILOptionalModule() as imp:
            Image = imp("Image")
            im = Image.open(imageFile)
            return im.size

    @staticmethod
    def generatePreviewFromImage(image):
        fd, path = tempfile.mkstemp()

        preview = None
        if ImageTools.scaleImage(image, path, "JPEG", YowConstants.PREVIEW_WIDTH, YowConstants.PREVIEW_HEIGHT):
            fileObj = os.fdopen(fd, "rb+")
            fileObj.seek(0)
            preview = fileObj.read()
            fileObj.close()
        os.remove(path)
        return preview

class MimeTools:
    MIME_FILE = os.path.join(os.path.dirname(__file__), 'mime.types')
    mimetypes.init() # Load default mime.types
    try:
        mimetypes.init([MIME_FILE]) # Append whatsapp mime.types
    except exception as e:
        logger.warning("Mime types supported can't be read. System mimes will be used. Cause: " + e.message)

    @staticmethod
    def getMIME(filepath):
        mimeType = mimetypes.guess_type(filepath)[0]
        if mimeType is None:
            raise Exception("Unsupported/unrecognized file type for: "+filepath);
        return mimeType

class VideoTools:
    @staticmethod
    def getVideoProperties(videoFile):
        with FFVideoOptionalModule() as imp:
            VideoStream = imp("VideoStream")
            s = VideoStream(videoFile)
            return s.width, s.height, s.bitrate, s.duration #, s.codec_name

    @staticmethod
    def generatePreviewFromVideo(videoFile):
        with FFVideoOptionalModule() as imp:
            VideoStream = imp("VideoStream")
            fd, path = tempfile.mkstemp('.jpg')
            stream = VideoStream(videoFile)
            stream.get_frame_at_sec(0).image().save(path)
            preview = ImageTools.generatePreviewFromImage(path)
            os.remove(path)
            return preview



from yowsup.layers.network.dispatcher.dispatcher import YowConnectionDispatcher
import asyncore
import logging
import socket
import traceback

logger = logging.getLogger(__name__)


class AsyncoreConnectionDispatcher(YowConnectionDispatcher, asyncore.dispatcher_with_send):
    def __init__(self, connectionCallbacks):
        super(AsyncoreConnectionDispatcher, self).__init__(connectionCallbacks)
        asyncore.dispatcher_with_send.__init__(self)
        self._connected = False

    def sendData(self, data):
        if self._connected:
            self.out_buffer = self.out_buffer + data
            self.initiate_send()
        else:
            logger.warn("Attempted to send %d bytes while still not connected" % len(data))

    def connect(self, host):
        logger.debug("connect(%s)" % str(host))
        self.connectionCallbacks.onConnecting()
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        asyncore.dispatcher_with_send.connect(self, host)
        asyncore.loop(timeout=1)

    def handle_connect(self):
        logger.debug("handle_connect")
        if not self._connected:
            self._connected = True
            self.connectionCallbacks.onConnected()

    def handle_close(self):
        logger.debug("handle_close")
        self.close()
        self._connected = False
        self.connectionCallbacks.onDisconnected()

    def handle_error(self):
        logger.error(traceback.format_exc())
        self.handle_close()

    def handle_read(self):
        data = self.recv(1024)
        self.connectionCallbacks.onRecvData(data)

    def disconnect(self):
        logger.debug("disconnect")
        self.handle_close()



from yowsup.layers import YowParallelLayer
import time, logging, random
from yowsup.layers import YowLayer
from yowsup.layers.noise.layer import YowNoiseLayer
from yowsup.layers.noise.layer_noise_segments import YowNoiseSegmentsLayer
from yowsup.layers.auth                        import YowAuthenticationProtocolLayer
from yowsup.layers.coder                       import YowCoderLayer
from yowsup.layers.logger                      import YowLoggerLayer
from yowsup.layers.network                     import YowNetworkLayer
from yowsup.layers.protocol_messages           import YowMessagesProtocolLayer
from yowsup.layers.protocol_media              import YowMediaProtocolLayer
from yowsup.layers.protocol_acks               import YowAckProtocolLayer
from yowsup.layers.protocol_receipts           import YowReceiptProtocolLayer
from yowsup.layers.protocol_groups             import YowGroupsProtocolLayer
from yowsup.layers.protocol_presence           import YowPresenceProtocolLayer
from yowsup.layers.protocol_ib                 import YowIbProtocolLayer
from yowsup.layers.protocol_notifications      import YowNotificationsProtocolLayer
from yowsup.layers.protocol_iq                 import YowIqProtocolLayer
from yowsup.layers.protocol_contacts           import YowContactsIqProtocolLayer
from yowsup.layers.protocol_chatstate          import YowChatstateProtocolLayer
from yowsup.layers.protocol_privacy            import YowPrivacyProtocolLayer
from yowsup.layers.protocol_profiles           import YowProfilesProtocolLayer
from yowsup.layers.protocol_calls import YowCallsProtocolLayer
from yowsup.env import YowsupEnv
from yowsup.common.constants import YowConstants
from yowsup.layers.axolotl import AxolotlSendLayer, AxolotlControlLayer, AxolotlReceivelayer
import inspect
try:
    import Queue
except ImportError:
    import queue as Queue
logger = logging.getLogger(__name__)

YOWSUP_PROTOCOL_LAYERS_BASIC = (
    YowAuthenticationProtocolLayer, YowMessagesProtocolLayer,
    YowReceiptProtocolLayer, YowAckProtocolLayer, YowPresenceProtocolLayer,
    YowIbProtocolLayer, YowIqProtocolLayer, YowNotificationsProtocolLayer,
    YowContactsIqProtocolLayer, YowChatstateProtocolLayer, YowCallsProtocolLayer

)


class YowStackBuilder(object):
    def __init__(self):
        self.layers = ()
        self._props = {}

    def setProp(self, key, value):
        self._props[key] = value
        return self

    def pushDefaultLayers(self):
        defaultLayers = YowStackBuilder.getDefaultLayers()
        self.layers += defaultLayers
        return self

    def push(self, yowLayer):
        self.layers += (yowLayer,)
        return self

    def pop(self):
        self.layers = self.layers[:-1]
        return self

    def build(self):
        return YowStack(self.layers, reversed = False, props = self._props)

    @staticmethod
    def getDefaultLayers(groups = True, media = True, privacy = True, profiles = True):
        coreLayers = YowStackBuilder.getCoreLayers()
        protocolLayers = YowStackBuilder.getProtocolLayers(groups = groups, media=media, privacy=privacy, profiles=profiles)

        allLayers = coreLayers
        allLayers += (AxolotlControlLayer,)
        allLayers += (YowParallelLayer((AxolotlSendLayer, AxolotlReceivelayer)),)

        allLayers += (YowParallelLayer(protocolLayers),)

        return allLayers

    @staticmethod
    def getDefaultStack(layer = None, axolotl = False, groups = True, media = True, privacy = True, profiles = True):
        """
        :param layer: An optional layer to put on top of default stack
        :param axolotl: E2E encryption enabled/ disabled
        :return: YowStack
        """

        allLayers = YowStackBuilder.getDefaultLayers(axolotl, groups = groups, media=media,privacy=privacy, profiles=profiles)
        if layer:
            allLayers = allLayers + (layer,)


        return YowStack(allLayers, reversed = False)

    @staticmethod
    def getCoreLayers():
        return (
            YowLoggerLayer,
            YowCoderLayer,
            YowNoiseLayer,
            YowNoiseSegmentsLayer,
            YowNetworkLayer
        )[::-1]

    @staticmethod
    def getProtocolLayers(groups = True, media = True, privacy = True, profiles = True):
        layers = YOWSUP_PROTOCOL_LAYERS_BASIC
        if groups:
            layers += (YowGroupsProtocolLayer,)

        if media:
            layers += (YowMediaProtocolLayer, )

        if privacy:
            layers += (YowPrivacyProtocolLayer, )

        if profiles:
            layers += (YowProfilesProtocolLayer, )

        return layers

class YowStack(object):
    __stack = []
    __stackInstances = []
    __detachedQueue = Queue.Queue()
    def __init__(self, stackClassesArr = None, reversed = True, props = None):
        stackClassesArr = stackClassesArr or ()
        self.__stack = stackClassesArr[::-1] if reversed else stackClassesArr
        self.__stackInstances = []
        self._props = props or {}

        self.setProp(YowNetworkLayer.PROP_ENDPOINT, YowConstants.ENDPOINTS[random.randint(0,len(YowConstants.ENDPOINTS)-1)])
        self._construct()


    def getLayerInterface(self, YowLayerClass):
        for inst in self.__stackInstances:
            if inst.__class__ == YowLayerClass:
                return inst.getLayerInterface()
            elif inst.__class__ == YowParallelLayer:
                res = inst.getLayerInterface(YowLayerClass)
                if res:
                    return res


    def send(self, data):
        self.__stackInstances[-1].send(data)

    def receive(self, data):
        self.__stackInstances[0].receive(data)

    def setCredentials(self, credentials):
        self.getLayerInterface(YowAuthenticationProtocolLayer).setCredentials(*credentials)

    def addLayer(self, layerClass):
        self.__stack.push(layerClass)

    def addPostConstructLayer(self, layer):
        self.__stackInstances[-1].setLayers(layer, self.__stackInstances[-2])
        layer.setLayers(None, self.__stackInstances[-1])
        self.__stackInstances.append(layer)

    def setProp(self, key, value):
        self._props[key] = value

    def getProp(self, key, default = None):
        return self._props[key] if key in self._props else default

    def emitEvent(self, yowLayerEvent):
        if not self.__stackInstances[0].onEvent(yowLayerEvent):
            self.__stackInstances[0].emitEvent(yowLayerEvent)

    def broadcastEvent(self, yowLayerEvent):
        if not self.__stackInstances[-1].onEvent(yowLayerEvent):
            self.__stackInstances[-1].broadcastEvent(yowLayerEvent)

    def execDetached(self, fn):
        self.__class__.__detachedQueue.put(fn)

    def loop(self, *args, **kwargs):
        while True:
            try:
                callback = self.__class__.__detachedQueue.get(False) #doesn't block
                callback()
            except Queue.Empty:
                pass
            time.sleep(0.1)

    def _construct(self):
        logger.debug("Initializing stack")
        for s in self.__stack:
            if type(s) is tuple:
                logger.warn("Implicit declaration of parallel layers in a tuple is deprecated, pass a YowParallelLayer instead")
                inst = YowParallelLayer(s)
            else:
                if inspect.isclass(s):
                    if issubclass(s, YowLayer):
                        inst = s()
                    else:
                        raise ValueError("Stack must contain only subclasses of YowLayer")
                elif issubclass(s.__class__, YowLayer):
                        inst = s
                else:
                    raise ValueError("Stack must contain only subclasses of YowLayer")
                #inst = s()
            logger.debug("Constructed %s" % inst)
            inst.setStack(self)
            self.__stackInstances.append(inst)

        for i in range(0, len(self.__stackInstances)):
            upperLayer = self.__stackInstances[i + 1] if (i + 1) < len(self.__stackInstances) else None
            lowerLayer = self.__stackInstances[i - 1] if i > 0 else None
            self.__stackInstances[i].setLayers(upperLayer, lowerLayer)

    def getLayer(self, layerIndex):
        return self.__stackInstances[layerIndex]



from yowsup.layers.protocol_media.mediacipher import MediaCipher
import base64
import unittest


class MediaCipherTest(unittest.TestCase):
    IMAGE = (
        b'EOnnZIBu1vTSI51IeJvaKR+8W1FqBETATI2Ikl6nVQ8=',

        b'/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAMCAgICAgMCAgIDAwMDBAYEBAQEBAgGBgUGCQgKCgkICQkKDA8M'
        b'CgsOCwkJDRENDg8QEBEQCgwSExIQEw8QEBD/2wBDAQMDAwQDBAgEBAgQCwkLEBAQEBAQEBAQEBAQEBAQEBAQ'
        b'EBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBD/wAARCAABAAEDAREAAhEBAxEB/8QAHwAAAQUBAQEB'
        b'AQEAAAAAAAAAAAECAwQFBgcICQoL/8QAtRAAAgEDAwIEAwUFBAQAAAF9AQIDAAQRBRIhMUEGE1FhByJxFDKB'
        b'kaEII0KxwRVS0fAkM2JyggkKFhcYGRolJicoKSo0NTY3ODk6Q0RFRkdISUpTVFVWV1hZWmNkZWZnaGlqc3R1'
        b'dnd4eXqDhIWGh4iJipKTlJWWl5iZmqKjpKWmp6ipqrKztLW2t7i5usLDxMXGx8jJytLT1NXW19jZ2uHi4+Tl'
        b'5ufo6erx8vP09fb3+Pn6/8QAHwEAAwEBAQEBAQEBAQAAAAAAAAECAwQFBgcICQoL/8QAtREAAgECBAQDBAcF'
        b'BAQAAQJ3AAECAxEEBSExBhJBUQdhcRMiMoEIFEKRobHBCSMzUvAVYnLRChYkNOEl8RcYGRomJygpKjU2Nzg5'
        b'OkNERUZHSElKU1RVVldYWVpjZGVmZ2hpanN0dXZ3eHl6goOEhYaHiImKkpOUlZaXmJmaoqOkpaanqKmqsrO0'
        b'tba3uLm6wsPExcbHyMnK0tPU1dbX2Nna4uPk5ebn6Onq8vP09fb3+Pn6/9oADAMBAAIRAxEAPwD9U6AP/9k'
        b'==',

        b'dT9YPFSz4dprkH6uiXPEVERGZVIZbGYHzwue21WoLP1RCE2wmu11M8n6ysfROPtI39DCRFQhBBEVGFCT/nfV'
        b'pt+fouIENBSXY44mR2en4HGvRR//dlM5OBLz2WuEOf01iKPazGtfacy6lnV0X5JagL4r1mKeyuSXJEV81kxj'
        b'Vd3OArpLKt13XM36PcnTd/U6DHOV6Vf982Wc1UjR7kMb5JT+HlWrvz9CCGMTX5mqBYWEr3InCFyrmaZu8DXC'
        b'60YUZCPLTHJP0hFQA1ooKQks4f3F39tzVL3dbX3io7XPQiSgHN6nuPCD0PF7dWINep+amk+ODjeQd/o2guqx'
        b'O/AngNIxfFWq3915jMQWAXeARUFw7x+9Qx93UWC/sfQ72nTqdQaH5W4vsMUKaocZcAJ7YWudKo5Y15uo3ulP'
        b'744Cyo54tvxUSV0zLC90LYCe8fJefREjreO73RlVqoynTyFHuVS7/YMnesO5lCHZJ2NpHuzpGKCU08RgCuSr'
        b'K/wh4SLXZ1nhZEYX/xY4cDO7swrA3Y3Qt0gjFzNBfUT9aABQaU+WHY8pBS/oVfJsuj2iod2fsW8UoplImoOk'
        b'bosYlMkdZSIfwBQ5kt3On2d+HPUNt7HVZWRECFj0C4IHhZCZIJw0wFmPMOiIHyzittXGb45uJA2Sd1gnRbw0'
        b'1XFuoIyvS7z0MtW0QUiD6kaJFBkpTo7hxN/HyN8xQwOkeBcKORdpeSp62iPBuRvel9I2p07it7UyYhzas+Jv'
        b'tl6i+hIz6Z5nzQEZ+zPAYxDmoy4h9GQuXUivTzU9I0/Sd4QvM3rfewGeXsNSjWI1CLVZG+4wL1TPCbQGaHEB'
        b'NF8zmpTMYV+NvCzv/2DwYuYoiUI='
    )

    def setUp(self):
        self._cipher = MediaCipher()  # type: MediaCipher

    def test_decrypt_image(self):
        media_key, media_plaintext, media_ciphertext = map(base64.b64decode, self.IMAGE)
        self.assertEqual(media_plaintext, self._cipher.decrypt_image(media_ciphertext, media_key))

    def test_encrypt_image(self):
        media_key, media_plaintext, media_ciphertext = map(base64.b64decode, self.IMAGE)
        encrypted = self._cipher.encrypt(media_plaintext, media_key, MediaCipher.INFO_IMAGE)
        self.assertEqual(media_ciphertext, encrypted)



class ConfigSerialize(object):

    def __init__(self, transforms):
        self._transforms = transforms

    def serialize(self, config):
        """
        :param config:
        :type config: yowsup.config.base.config.Config
        :return:
        :rtype: bytes
        """
        for transform in self._transforms:
            config = transform.transform(config)
        return config

    def deserialize(self, data):
        """
        :type cls: type
        :param data:
        :type data: bytes
        :return:
        :rtype: yowsup.config.base.config.Config
        """
        for transform in self._transforms[::-1]:
            data = transform.reverse(data)
        return data



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class SuccessProtocolEntity(ProtocolEntity):
    def __init__(self, creation, props, t, location):
        super(SuccessProtocolEntity, self).__init__("success")
        self.location = location
        self.creation = int(creation)
        self.props = props
        self.t = int(t) ##whatever that is !

    def __str__(self):
        out  = "Account:\n"
        out += "Location: %s\n" % self.location
        out += "Creation: %s\n" % self.creation
        out += "Props: %s\n" % self.props
        out += "t: %s\n" % self.t
        return out

    def toProtocolTreeNode(self):
        attributes = {
            "location"      :  self.location,
            "creation"    :    str(self.creation),
            "props"       :    self.props,
            "t"           :    str(self.t)
        }
        return self._createProtocolTreeNode(attributes)

    @staticmethod
    def fromProtocolTreeNode(node):
        return SuccessProtocolEntity(
            node.getAttributeValue("creation"),
            node.getAttributeValue("props"),
            node.getAttributeValue("t"),
            node.getAttributeValue("location")
            )


from axolotl.util.keyhelper import KeyHelper
from axolotl.identitykeypair import IdentityKeyPair
from axolotl.groups.senderkeyname import SenderKeyName
from axolotl.axolotladdress import AxolotlAddress
from axolotl.sessioncipher import SessionCipher
from axolotl.groups.groupcipher import GroupCipher
from axolotl.groups.groupsessionbuilder import GroupSessionBuilder
from axolotl.sessionbuilder import SessionBuilder
from axolotl.protocol.prekeywhispermessage import PreKeyWhisperMessage
from axolotl.protocol.whispermessage import WhisperMessage
from axolotl.state.prekeybundle import PreKeyBundle
from axolotl.untrustedidentityexception import UntrustedIdentityException
from axolotl.invalidmessageexception import InvalidMessageException
from axolotl.duplicatemessagexception import DuplicateMessageException
from axolotl.invalidkeyidexception import InvalidKeyIdException
from axolotl.nosessionexception import NoSessionException
from axolotl.protocol.senderkeydistributionmessage import SenderKeyDistributionMessage
from axolotl.state.axolotlstore import AxolotlStore
from yowsup.axolotl.store.sqlite.liteaxolotlstore import LiteAxolotlStore
from yowsup.axolotl import exceptions
import random
import logging
import sys

logger = logging.getLogger(__name__)


class AxolotlManager(object):

    COUNT_GEN_PREKEYS = 812
    THRESHOLD_REGEN = 10
    MAX_SIGNED_PREKEY_ID = 16777215

    def __init__(self, store, username):
        """
        :param store:
        :type store: AxolotlStore
        :param username:
        :type username: str
        """
        self._username = username # type: str
        self._store = store # type: LiteAxolotlStore
        self._identity = self._store.getIdentityKeyPair() # type: IdentityKeyPair
        self._registration_id = self._store.getLocalRegistrationId() # type: int | None

        assert self._registration_id is not None
        assert self._identity is not None

        self._group_session_builder = GroupSessionBuilder(self._store) # type: GroupSessionBuilder
        self._session_ciphers = {} # type: dict[str, SessionCipher]
        self._group_ciphers = {} # type: dict[str, GroupCipher]
        logger.debug("Initialized AxolotlManager [username=%s, db=%s]" % (self._username, store))

    @property
    def registration_id(self):
        return self._registration_id

    @property
    def identity(self):
        return self._identity

    def level_prekeys(self, force=False):
        logger.debug("level_prekeys(force=%s)" % force)
        pending_prekeys = self._store.loadPreKeys()
        logger.debug("len(pending_prekeys) = %d" % len(pending_prekeys))
        if force or len(pending_prekeys) < self.THRESHOLD_REGEN:
            count_gen = self.COUNT_GEN_PREKEYS - len(pending_prekeys)
            logger.info("Generating %d prekeys" % count_gen)
            ## arbitrary, should keep track of generated prekey ids and create from there
            prekeys = KeyHelper.generatePreKeys(KeyHelper.getRandomSequence(2**32 // 2), count_gen)
            logger.info("Storing %d prekeys" % len(prekeys))
            for i in range(0, len(prekeys)):
                key = prekeys[i]
                if logger.level <= logging.DEBUG:
                    sys.stdout.write("Storing prekey %d/%d \r" % (i + 1, len(prekeys)))
                    sys.stdout.flush()
                self._store.storePreKey(key.getId(), key)
            return prekeys
        return []

    def load_unsent_prekeys(self):
        logger.debug("load_unsent_prekeys")
        unsent = self._store.preKeyStore.loadUnsentPendingPreKeys()
        if len(unsent) > 0:
            logger.info("Loaded %d unsent prekeys" % len(unsent))
        return unsent

    def set_prekeys_as_sent(self, prekeyIds):
        """
        :param prekeyIds:
        :type prekeyIds: list
        :return:
        :rtype:
        """
        logger.debug("set_prekeys_as_sent(prekeyIds=[%d prekeyIds])" % len(prekeyIds))
        self._store.preKeyStore.setAsSent([prekey.getId() for prekey in prekeyIds])

    def generate_signed_prekey(self):
        logger.debug("generate_signed_prekey")
        latest_signed_prekey = self.load_latest_signed_prekey(generate=False)
        if latest_signed_prekey is not None:
            if latest_signed_prekey.getId() == self.MAX_SIGNED_PREKEY_ID:
                new_signed_prekey_id = (self.MAX_SIGNED_PREKEY_ID / 2) + 1
            else:
                new_signed_prekey_id = latest_signed_prekey.getId() + 1
        else:
            new_signed_prekey_id = 0
        signed_prekey = KeyHelper.generateSignedPreKey(self._identity, new_signed_prekey_id)
        self._store.storeSignedPreKey(signed_prekey.getId(), signed_prekey)
        return signed_prekey

    def load_latest_signed_prekey(self, generate=False):
        logger.debug("load_latest_signed_prekey")
        signed_prekeys = self._store.loadSignedPreKeys()
        if len(signed_prekeys):
            return signed_prekeys[-1]

        return self.generate_signed_prekey() if generate else None

    def _get_session_cipher(self, recipientid):
        logger.debug("get_session_cipher(recipientid=%s)" % recipientid)
        if recipientid in self._session_ciphers:
            session_cipher = self._session_ciphers[recipientid]
        else:
            session_cipher= SessionCipher(self._store, self._store, self._store, self._store, recipientid, 1)
            self._session_ciphers[recipientid] = session_cipher
        return session_cipher

    def _get_group_cipher(self, groupid, username):
        logger.debug("get_group_cipher(groupid=%s, username=%s)" % (groupid, username))
        senderkeyname = SenderKeyName(groupid, AxolotlAddress(username, 0))
        if senderkeyname in self._group_ciphers:
            group_cipher = self._group_ciphers[senderkeyname]
        else:
            group_cipher = GroupCipher(self._store.senderKeyStore, senderkeyname)
            self._group_ciphers[senderkeyname] = group_cipher
        return group_cipher

    def _generate_random_padding(self):
        logger.debug("generate_random_padding")
        num = random.randint(1,255)
        return bytes(bytearray([num] * num))

    def _unpad(self, data):
        padding_byte = data[-1] if type(data[-1]) is int else ord(data[-1]) # bec inconsistent API?
        padding = padding_byte & 0xFF
        return data[:-padding]

    def encrypt(self, recipient_id, message):
        logger.debug("encrypt(recipientid=%s, message=%s)" % (recipient_id, message))
        """
        :param recipient_id:
        :type recipient_id: str
        :param data:
        :type data: bytes
        :return:
        :rtype:
        """
        cipher = self._get_session_cipher(recipient_id)
        return cipher.encrypt(message + self._generate_random_padding())

    def decrypt_pkmsg(self, senderid, data, unpad):
        logger.debug("decrypt_pkmsg(senderid=%s, data=(omitted), unpad=%s)" % (senderid, unpad))
        pkmsg = PreKeyWhisperMessage(serialized=data)
        try:
            plaintext = self._get_session_cipher(senderid).decryptPkmsg(pkmsg)
            return self._unpad(plaintext) if unpad else plaintext
        except NoSessionException:
            raise exceptions.NoSessionException()
        except InvalidKeyIdException:
            raise exceptions.InvalidKeyIdException()
        except InvalidMessageException:
            raise exceptions.InvalidMessageException()
        except DuplicateMessageException:
            raise exceptions.DuplicateMessageException()


    def decrypt_msg(self, senderid, data, unpad):
        logger.debug("decrypt_msg(senderid=%s, data=[omitted], unpad=%s)" % (senderid, unpad))
        msg = WhisperMessage(serialized=data)
        try:
            plaintext = self._get_session_cipher(senderid).decryptMsg(msg)

            return self._unpad(plaintext) if unpad else plaintext
        except NoSessionException:
            raise exceptions.NoSessionException()
        except InvalidKeyIdException:
            raise exceptions.InvalidKeyIdException()
        except InvalidMessageException:
            raise exceptions.InvalidMessageException()
        except DuplicateMessageException:
            raise exceptions.DuplicateMessageException()

    def group_encrypt(self, groupid, message):
        """
        :param groupid:
        :type groupid: str
        :param message:
        :type message: bytes
        :return:
        :rtype:
        """
        logger.debug("group_encrypt(groupid=%s, message=%s)" % (groupid, message))
        group_cipher = self._get_group_cipher(groupid, self._username)
        return group_cipher.encrypt(message + self._generate_random_padding())

    def group_decrypt(self, groupid, participantid, data):
        logger.debug("group_decrypt(groupid=%s, participantid=%s, data=[omitted])" % (groupid, participantid))
        group_cipher = self._get_group_cipher(groupid, participantid)
        try:
            plaintext = group_cipher.decrypt(data)
            plaintext = self._unpad(plaintext)
            return plaintext
        except NoSessionException:
            raise exceptions.NoSessionException()
        except DuplicateMessageException:
            raise exceptions.DuplicateMessageException()

    def group_create_skmsg(self, groupid):
        logger.debug("group_create_skmsg(groupid=%s)" % groupid)
        senderKeyName = SenderKeyName(groupid, AxolotlAddress(self._username, 0))
        return self._group_session_builder.create(senderKeyName)

    def group_create_session(self, groupid, participantid, skmsgdata):
        """
        :param groupid:
        :type groupid: str
        :param participantid:
        :type participantid: str
        :param skmsgdata:
        :type skmsgdata: bytearray
        :return:
        :rtype:
        """
        logger.debug("group_create_session(groupid=%s, participantid=%s, skmsgdata=[omitted])"
                     % (groupid, participantid))
        senderKeyName = SenderKeyName(groupid, AxolotlAddress(participantid, 0))
        senderkeydistributionmessage = SenderKeyDistributionMessage(serialized=skmsgdata)
        self._group_session_builder.process(senderKeyName, senderkeydistributionmessage)

    def create_session(self, username, prekeybundle, autotrust=False):
        """
        :param username:
        :type username: str
        :param prekeybundle:
        :type prekeybundle: PreKeyBundle
        :return:
        :rtype:
        """
        logger.debug("create_session(username=%s, prekeybundle=[omitted], autotrust=%s)" % (username, autotrust))
        session_builder = SessionBuilder(self._store, self._store, self._store, self._store, username, 1)
        try:
            session_builder.processPreKeyBundle(prekeybundle)
        except UntrustedIdentityException as ex:
            if autotrust:
                self.trust_identity(ex.getName(), ex.getIdentityKey())
            else:
                raise exceptions.UntrustedIdentityException(ex.getName(), ex.getIdentityKey())

    def session_exists(self, username):
        """
        :param username:
        :type username: str
        :return:
        :rtype:
        """
        logger.debug("session_exists(%s)?" % username)
        return self._store.containsSession(username, 1)


    def load_senderkey(self, groupid):
        logger.debug("load_senderkey(groupid=%s)" % groupid)
        senderkeyname = SenderKeyName(groupid, AxolotlAddress(self._username, 0))
        return self._store.loadSenderKey(senderkeyname)

    def trust_identity(self, recipientid, identitykey):
        logger.debug("trust_identity(recipientid=%s, identitykey=[omitted])" % recipientid)
        self._store.saveIdentity(recipientid, identitykey)



import abc
import logging
from six import with_metaclass

logger = logging.getLogger(__name__)

DEFAULT = "android"


class YowsupEnvType(abc.ABCMeta):
    def __init__(cls, name, bases, dct):
        if name != "YowsupEnv":
            YowsupEnv.registerEnv(cls)
        super(YowsupEnvType, cls).__init__(name, bases, dct)


class YowsupEnv(with_metaclass(YowsupEnvType, object)):
    __metaclass__ = YowsupEnvType
    __ENVS = {}
    __CURR = None

    _USERAGENT_STRING = "WhatsApp/{WHATSAPP_VERSION} {OS_NAME}/{OS_VERSION} Device/{MANUFACTURER}-{DEVICE_NAME}"

    @classmethod
    def registerEnv(cls, envCls):
        envName = envCls.__name__.lower().replace("yowsupenv", "")
        cls.__ENVS[envName] = envCls
        logger.debug("registered env %s => %s" % (envName, envCls))

    @classmethod
    def setEnv(cls, envName):
        if not envName in cls.__ENVS:
            raise ValueError("%s env does not exist" % envName)
        logger.debug("Current env changed to %s " % envName)
        cls.__CURR = cls.__ENVS[envName]()

    @classmethod
    def getEnv(cls, envName):
        if not envName in cls.__ENVS:
            raise ValueError("%s env does not exist" % envName)

        return cls.__ENVS[envName]()

    @classmethod
    def getRegisteredEnvs(cls):
        return list(cls.__ENVS.keys())

    @classmethod
    def getCurrent(cls):
        """
        :rtype: YowsupEnv
        """
        if cls.__CURR is None:
            env = DEFAULT
            envs = cls.getRegisteredEnvs()
            if env not in envs:
                env = envs[0]
            logger.debug("Env not set, setting it to %s" % env)
            cls.setEnv(env)
        return cls.__CURR

    @abc.abstractmethod
    def getToken(self, phoneNumber):
        pass

    @abc.abstractmethod
    def getVersion(self):
        pass

    @abc.abstractmethod
    def getOSVersion(self):
        pass

    @abc.abstractmethod
    def getOSName(self):
        pass

    @abc.abstractmethod
    def getDeviceName(self):
        pass

    @abc.abstractmethod
    def getManufacturer(self):
        pass

    def getBuildVersion(self):
        pass

    def getUserAgent(self):
        return self.__class__._USERAGENT_STRING.format(
            WHATSAPP_VERSION=self.getVersion(),
            OS_NAME=self.getOSName(),
            OS_VERSION=self.getOSVersion(),
            MANUFACTURER=self.getManufacturer(),
            DEVICE_NAME=self.getDeviceName()
        )



from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
from yowsup.env import YowsupEnv


class WAExistsRequest(WARequest):

    def __init__(self, config):
        """
        :param config:
        :type config: yowsup.config.v1.config.Config
        """
        super(WAExistsRequest,self).__init__(config)
        if config.id is None:
            raise ValueError("Config does not contain id")

        self.url = "v.whatsapp.net/v2/exist"

        self.pvars = ["status", "reason", "sms_length", "voice_length", "result","param", "login", "type",
                      "chat_dns_domain", "edge_routing_info"
                    ]

        self.setParser(JSONResponseParser())
        self.addParam("token", YowsupEnv.getCurrent().getToken(self._p_in))



from .layer_base import AxolotlBaseLayer

from yowsup.layers.protocol_receipts.protocolentities import OutgoingReceiptProtocolEntity
from yowsup.layers.protocol_messages.proto.e2e_pb2 import *
from yowsup.layers.axolotl.protocolentities import *
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_messages.protocolentities.proto import ProtoProtocolEntity
from yowsup.layers.axolotl.props import PROP_IDENTITY_AUTOTRUST
from yowsup.axolotl import exceptions

from axolotl.untrustedidentityexception import UntrustedIdentityException

import logging
logger = logging.getLogger(__name__)


class AxolotlReceivelayer(AxolotlBaseLayer):
    def __init__(self):
        super(AxolotlReceivelayer, self).__init__()
        self.v2Jids = [] #people we're going to send v2 enc messages
        self.sessionCiphers = {}
        self.groupCiphers = {}
        self.pendingIncomingMessages = {} #(jid, participantJid?) => message

    def receive(self, protocolTreeNode):
        """
        :type protocolTreeNode: ProtocolTreeNode
        """
        if not self.processIqRegistry(protocolTreeNode):
            if protocolTreeNode.tag == "message":
                self.onMessage(protocolTreeNode)
            elif not protocolTreeNode.tag == "receipt":
                #receipts will be handled by send layer
                self.toUpper(protocolTreeNode)

            # elif protocolTreeNode.tag == "iq":
            #     if protocolTreeNode.getChild("encr_media"):
            #         protocolTreeNode.addChild("media", {
            #             "url": protocolTreeNode["url"],
            #             "ip": protocolTreeNode["ip"],
            #         })
            #         self.toUpper(protocolTreeNode)
            #         return

    ######

    def onEncrMediaResult(self, resultNode):
        pass

    def processPendingIncomingMessages(self, jid, participantJid = None):
        conversationIdentifier = (jid, participantJid)
        if conversationIdentifier in self.pendingIncomingMessages:
            for messageNode in self.pendingIncomingMessages[conversationIdentifier]:
                self.onMessage(messageNode)

            del self.pendingIncomingMessages[conversationIdentifier]

    ##### handling received data #####

    def onMessage(self, protocolTreeNode):
        encNode = protocolTreeNode.getChild("enc")
        if encNode:
            self.handleEncMessage(protocolTreeNode)
        else:
            self.toUpper(protocolTreeNode)

    def handleEncMessage(self, node):
        encMessageProtocolEntity = EncryptedMessageProtocolEntity.fromProtocolTreeNode(node)
        isGroup =  node["participant"] is not None
        senderJid = node["participant"] if isGroup else node["from"]
        if node.getChild("enc")["v"] == "2" and node["from"] not in self.v2Jids:
            self.v2Jids.append(node["from"])
        try:
            if encMessageProtocolEntity.getEnc(EncProtocolEntity.TYPE_PKMSG):
                self.handlePreKeyWhisperMessage(node)
            elif encMessageProtocolEntity.getEnc(EncProtocolEntity.TYPE_MSG):
                self.handleWhisperMessage(node)
            if encMessageProtocolEntity.getEnc(EncProtocolEntity.TYPE_SKMSG):
                self.handleSenderKeyMessage(node)
        except (exceptions.InvalidMessageException, exceptions.InvalidKeyIdException) as e:
            logger.warning("InvalidMessage or KeyId for %s, going to send a retry", encMessageProtocolEntity.getAuthor(False))
            retry = RetryOutgoingReceiptProtocolEntity.fromMessageNode(node, self.manager.registration_id)
            self.toLower(retry.toProtocolTreeNode())
        except exceptions.NoSessionException:
            logger.warning("No session for %s, getting their keys now", encMessageProtocolEntity.getAuthor(False))

            conversationIdentifier = (node["from"], node["participant"])

            if conversationIdentifier not in self.pendingIncomingMessages:
                self.pendingIncomingMessages[conversationIdentifier] = []
            self.pendingIncomingMessages[conversationIdentifier].append(node)

            successFn = lambda successJids, b: self.processPendingIncomingMessages(*conversationIdentifier) if len(successJids) else None

            self.getKeysFor([senderJid], successFn)
        except exceptions.DuplicateMessageException:
            logger.warning("Received a message that we've previously decrypted, goint to send the delivery receipt myself")
            self.toLower(OutgoingReceiptProtocolEntity(node["id"], node["from"], participant=node["participant"]).toProtocolTreeNode())

        except UntrustedIdentityException as e:
            if self.getProp(PROP_IDENTITY_AUTOTRUST, False):
                logger.warning("Autotrusting identity for %s", e.getName())
                self.manager.trust_identity(e.getName(), e.getIdentityKey())
                return self.handleEncMessage(node)
            else:
                logger.error("Ignoring message with untrusted identity")

    def handlePreKeyWhisperMessage(self, node):
        pkMessageProtocolEntity = EncryptedMessageProtocolEntity.fromProtocolTreeNode(node)
        enc = pkMessageProtocolEntity.getEnc(EncProtocolEntity.TYPE_PKMSG)
        plaintext = self.manager.decrypt_pkmsg(pkMessageProtocolEntity.getAuthor(False), enc.getData(),
                                               enc.getVersion() == 2)

        if enc.getVersion() == 2:
            self.parseAndHandleMessageProto(pkMessageProtocolEntity, plaintext)

        node = pkMessageProtocolEntity.toProtocolTreeNode()
        node.addChild((ProtoProtocolEntity(plaintext, enc.getMediaType())).toProtocolTreeNode())

        self.toUpper(node)

    def handleWhisperMessage(self, node):
        encMessageProtocolEntity = EncryptedMessageProtocolEntity.fromProtocolTreeNode(node)

        enc = encMessageProtocolEntity.getEnc(EncProtocolEntity.TYPE_MSG)
        plaintext = self.manager.decrypt_msg(encMessageProtocolEntity.getAuthor(False), enc.getData(),
                                             enc.getVersion() == 2)

        if enc.getVersion() == 2:
            self.parseAndHandleMessageProto(encMessageProtocolEntity, plaintext)

        node = encMessageProtocolEntity.toProtocolTreeNode()
        node.addChild((ProtoProtocolEntity(plaintext, enc.getMediaType())).toProtocolTreeNode())

        self.toUpper(node)

    def handleSenderKeyMessage(self, node):
        encMessageProtocolEntity = EncryptedMessageProtocolEntity.fromProtocolTreeNode(node)
        enc = encMessageProtocolEntity.getEnc(EncProtocolEntity.TYPE_SKMSG)

        try:
            plaintext = self.manager.group_decrypt (
                groupid=encMessageProtocolEntity.getFrom(True),
                participantid=encMessageProtocolEntity.getParticipant(False),
                data=enc.getData()
            )
            self.parseAndHandleMessageProto(encMessageProtocolEntity, plaintext)

            node = encMessageProtocolEntity.toProtocolTreeNode()
            node.addChild((ProtoProtocolEntity(plaintext, enc.getMediaType())).toProtocolTreeNode())

            self.toUpper(node)

        except exceptions.NoSessionException:
            logger.warning("No session for %s, going to send a retry", encMessageProtocolEntity.getAuthor(False))
            retry = RetryOutgoingReceiptProtocolEntity.fromMessageNode(node, self.manager.registration_id)
            self.toLower(retry.toProtocolTreeNode())
        except exceptions.DuplicateMessageException:
            logger.warning(
                "Received a message that we've previously decrypted, goint to send the delivery receipt myself"
            )
            self.toLower(
                OutgoingReceiptProtocolEntity(
                    node["id"], node["from"], participant=node["participant"]
                ).toProtocolTreeNode()
            )

    def parseAndHandleMessageProto(self, encMessageProtocolEntity, serializedData):
        m = Message()
        try:
            m.ParseFromString(serializedData)
        except:
            print("DUMP:")
            print(serializedData)
            print([s for s in serializedData])
            # print([ord(s) for s in serializedData])
            raise
        if not m or not serializedData:
            raise ValueError("Empty message")

        if m.HasField("sender_key_distribution_message"):
            self.handleSenderKeyDistributionMessage(
                m.sender_key_distribution_message,
                encMessageProtocolEntity.getParticipant(False)
            )

    def handleSenderKeyDistributionMessage(self, senderKeyDistributionMessage, participantId):
        groupId = senderKeyDistributionMessage.group_id
        self.manager.group_create_session(
            groupid=groupId,
            participantid=participantId,
            skmsgdata=senderKeyDistributionMessage.axolotl_sender_key_distribution_message
        )





from yowsup.config.base.transform import ConfigTransform


class DictKeyValTransform(ConfigTransform):
    def transform(self, data):
        """
        :param data:
        :type data: dict
        :return:
        :rtype:
        """
        out=[]
        keys = sorted(data.keys())
        for k in keys:
            out.append("%s=%s" % (k, data[k]))
        return "\n".join(out)

    def reverse(self, data):
        out = {}
        for l in data.split('\n'):
            line = l.strip()
            if len(line) and line[0] not in ('#',';'):
                prep = line.split('#', 1)[0].split(';', 1)[0].split('=', 1)
                varname = prep[0].strip()
                val = prep[1].strip()
                out[varname.replace('-', '_')] = val
        return out



from yowsup.common.tools import ImageTools
import os


class ImageAttributes(object):
    def __init__(self, width, height, caption=None, jpeg_thumbnail=None):
        self.width = width
        self.height = height
        self.caption = caption
        self.jpeg_thumbnail = jpeg_thumbnail

    @staticmethod
    def from_filepath(filepath, dimensions=None, caption=None, jpeg_thumbnail=None):
        assert os.path.exists(filepath)
        if not jpeg_thumbnail:
            jpeg_thumbnail = ImageTools.generatePreviewFromImage(filepath)
        dimensions = dimensions or ImageTools.getImageDimensions(filepath)
        width, height = dimensions if dimensions else (None, None)
        assert width and height, "Could not determine image dimensions, install pillow or pass dimensions"

        return ImageAttributes(width, height, caption, jpeg_thumbnail)


from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_receipts.protocolentities import OutgoingReceiptProtocolEntity
from yowsup.layers.axolotl.protocolentities.iq_keys_get_result import ResultGetKeysIqProtocolEntity
class RetryOutgoingReceiptProtocolEntity(OutgoingReceiptProtocolEntity):

    '''
    <receipt type="retry" to="xxxxxxxxxxx@s.whatsapp.net" participant="" id="1415389947-12" t="1432833777">
        <retry count="1" t="1432833266" id="1415389947-12" v="1">
        </retry>
        <registration>
            HEX:xxxxxxxxx
        </registration>
    </receipt>
    '''

    def __init__(self, _id, jid, localRegistrationId, retryTimestamp, v = 1, count = 1, participant = None):
        super(RetryOutgoingReceiptProtocolEntity, self).__init__(_id, jid, participant=participant)
        '''
            Note to self: Android clients won't retry sending if the retry node didn't contain the message timestamp
        '''
        self.setRetryData(localRegistrationId, v,count, retryTimestamp)

    def setRetryData(self, localRegistrationId, v, count, retryTimestamp):
        self.localRegistrationId =  localRegistrationId
        self.v = v
        self.count = count
        self.retryTimestamp = int(retryTimestamp)

    def toProtocolTreeNode(self):
        node = super(RetryOutgoingReceiptProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("type", "retry")

        retryAttribs = {
            "count": str(self.count),
            "id":self.getId(),
            "v":str(self.v),
            "t":  str(self.retryTimestamp)
        }

        retry = ProtocolTreeNode("retry", retryAttribs)
        node.addChild(retry)
        registration = ProtocolTreeNode("registration", data=ResultGetKeysIqProtocolEntity._intToBytes(self.localRegistrationId))
        node.addChild(registration)
        return node

    def __str__(self):
        out = super(RetryOutgoingReceiptProtocolEntity, self).__str__()
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = OutgoingReceiptProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = RetryOutgoingReceiptProtocolEntity
        retryNode = node.getChild("retry")
        entity.setRetryData(ResultGetKeysIqProtocolEntity._bytesToInt(node.getChild("registration").data), retryNode["v"], retryNode["count"], retryNode["t"])

        return entity

    @staticmethod
    def fromMessageNode(messageNodeToBeRetried, localRegistrationId):
        return RetryOutgoingReceiptProtocolEntity(
            messageNodeToBeRetried.getAttributeValue("id"),
            messageNodeToBeRetried.getAttributeValue("from"),
            localRegistrationId,
            messageNodeToBeRetried.getAttributeValue("t"),
            participant=messageNodeToBeRetried.getAttributeValue("participant")
        )



class ConnectionCallbacks(object):
    def onConnected(self):
        pass

    def onDisconnected(self):
        pass

    def onRecvData(self, data):
        pass

    def onConnecting(self):
        pass

    def onConnectionError(self, error):
        pass


class YowConnectionDispatcher(object):
    def __init__(self, connectionCallbacks):
        assert isinstance(connectionCallbacks, ConnectionCallbacks)
        self.connectionCallbacks = connectionCallbacks

    def connect(self, host):
        pass

    def disconnect(self):
        pass

    def sendData(self, data):
        pass


from yowsup.stacks import  YowStackBuilder
from .layer import MediaSinkLayer
from yowsup.layers import YowLayerEvent
from yowsup.layers.network import YowNetworkLayer


class MediaSinkStack(object):
    def __init__(self, credentials, storage_dir=None):
        stackBuilder = YowStackBuilder()

        self._stack = stackBuilder\
            .pushDefaultLayers()\
            .push(MediaSinkLayer)\
            .build()
        self._stack.setProp(MediaSinkLayer.PROP_STORAGE_DIR, storage_dir)
        self._stack.setCredentials(credentials)

    def set_prop(self, key, val):
        self._stack.setProp(key, val)

    def start(self):
        self._stack.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))
        self._stack.loop()



from yowsup.structs import ProtocolEntity
from yowsup.layers.protocol_receipts.protocolentities  import OutgoingReceiptProtocolEntity
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes
from copy import deepcopy


class MessageProtocolEntity(ProtocolEntity):

    MESSAGE_TYPE_TEXT = "text"
    MESSAGE_TYPE_MEDIA = "media"

    def __init__(self, messageType, messageAttributes):
        """
        :type messageType: str
        :type messageAttributes: MessageAttributes
        """
        super(MessageProtocolEntity, self).__init__("message")
        assert type(messageAttributes) is MessageAttributes

        self._type = messageType
        self._id = messageAttributes.id or self._generateId()
        self._from = messageAttributes.sender
        self.to = messageAttributes.recipient
        self.timestamp = messageAttributes.timestamp or self._getCurrentTimestamp()
        self.notify = messageAttributes.notify
        self.offline = messageAttributes.offline
        self.retry = messageAttributes.retry
        self.participant= messageAttributes.participant

    def getType(self):
        return self._type

    def getId(self):
        return self._id

    def getTimestamp(self):
        return self.timestamp

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]

    def isBroadcast(self):
        return False

    def getTo(self, full = True):
        return self.to if full else self.to.split('@')[0]

    def getParticipant(self, full = True):
        return self.participant if full else self.participant.split('@')[0]

    def getAuthor(self, full = True):
        return self.getParticipant(full) if self.isGroupMessage() else self.getFrom(full)

    def getNotify(self):
        return self.notify

    def toProtocolTreeNode(self):
        attribs = {
            "type"      : self._type,
            "id"        : self._id,
        }

        if self.participant:
            attribs["participant"] = self.participant

        if self.isOutgoing():
            attribs["to"] = self.to
        else:
            attribs["from"] = self._from

            attribs["t"] = str(self.timestamp)

            if self.offline is not None:
               attribs["offline"] = "1" if self.offline else "0"
            if self.notify:
                attribs["notify"] = self.notify
            if self.retry:
                attribs["retry"] = str(self.retry)


        xNode = None
        #if self.isOutgoing():
        #    serverNode = ProtocolTreeNode("server", {})
        #    xNode = ProtocolTreeNode("x", {"xmlns": "jabber:x:event"}, [serverNode])


        return self._createProtocolTreeNode(attribs, children = [xNode] if xNode else None, data = None)

    def isOutgoing(self):
        return self._from is None

    def isGroupMessage(self):
        if self.isOutgoing():
            return "-" in self.to
        return self.participant != None

    def __str__(self):
        out  = "Message:\n"
        out += "ID: %s\n" % self._id
        out += "To: %s\n" % self.to  if self.isOutgoing() else "From: %s\n" % self._from
        out += "Type:  %s\n" % self._type
        out += "Timestamp: %s\n" % self.timestamp
        if self.participant:
            out += "Participant: %s\n" % self.participant
        return out

    def ack(self, read=False):
        return OutgoingReceiptProtocolEntity(self.getId(), self.getFrom(), read, participant=self.getParticipant())

    def forward(self, to, _id = None):
        OutgoingMessage = deepcopy(self)
        OutgoingMessage.to = to
        OutgoingMessage._from = None
        OutgoingMessage._id = self._generateId() if _id is None else _id
        return OutgoingMessage

    @staticmethod
    def fromProtocolTreeNode(node):
        return MessageProtocolEntity(
            node["type"],
            MessageAttributes.from_message_protocoltreenode(node)
        )



class NoSessionException(Exception):
    pass


class UntrustedIdentityException(Exception):
    def __init__(self, name, identity_key):
        self._name = name
        self._identity_key = identity_key

    @property
    def name(self):
        return self._name

    @property
    def identity_key(self):
        return self._identity_key


class InvalidMessageException(Exception):
    pass


class InvalidKeyIdException(Exception):
    pass


class DuplicateMessageException(Exception):
    pass




from yowsup.layers import YowLayer, YowLayerEvent, EventCallback
from yowsup.layers.network.layer_interface import YowNetworkLayerInterface
from yowsup.layers.network.dispatcher.dispatcher import ConnectionCallbacks
from yowsup.layers.network.dispatcher.dispatcher import YowConnectionDispatcher
from yowsup.layers.network.dispatcher.dispatcher_socket import SocketConnectionDispatcher
from yowsup.layers.network.dispatcher.dispatcher_asyncore import AsyncoreConnectionDispatcher
import logging
logger = logging.getLogger(__name__)


class YowNetworkLayer(YowLayer, ConnectionCallbacks):
    """This layer wraps a connection dispatcher that provides connection and a communication channel
    to remote endpoints. Unless explicitly configured, applications should not make assumption about
    the dispatcher being used as the default dispatcher could be changed across versions"""
    EVENT_STATE_CONNECT         = "org.openwhatsapp.yowsup.event.network.connect"
    EVENT_STATE_DISCONNECT      = "org.openwhatsapp.yowsup.event.network.disconnect"
    EVENT_STATE_CONNECTED       = "org.openwhatsapp.yowsup.event.network.connected"
    EVENT_STATE_DISCONNECTED    = "org.openwhatsapp.yowsup.event.network.disconnected"

    PROP_ENDPOINT               = "org.openwhatsapp.yowsup.prop.endpoint"
    PROP_NET_READSIZE           = "org.openwhatsapp.yowsup.prop.net.readSize"
    PROP_DISPATCHER             = "org.openwhatsapp.yowsup.prop.net.dispatcher"

    STATE_DISCONNECTED          = 0
    STATE_CONNECTING            = 1
    STATE_CONNECTED             = 2
    STATE_DISCONNECTING         = 3

    DISPATCHER_SOCKET = 0
    DISPATCHER_ASYNCORE = 1
    DISPATCHER_DEFAULT = DISPATCHER_ASYNCORE

    def __init__(self):
        self.state = self.__class__.STATE_DISCONNECTED
        YowLayer.__init__(self)
        ConnectionCallbacks.__init__(self)
        self.interface = YowNetworkLayerInterface(self)
        self.connected = False
        self._dispatcher = None  # type: YowConnectionDispatcher

    def __create_dispatcher(self, dispatcher_type):
        if dispatcher_type == self.DISPATCHER_ASYNCORE:
            logger.debug("Created asyncore dispatcher")
            return AsyncoreConnectionDispatcher(self)
        else:
            logger.debug("Created socket dispatcher")
            return SocketConnectionDispatcher(self)

    def onConnected(self):
        logger.debug("Connected")
        self.state = self.__class__.STATE_CONNECTED
        self.connected = True
        self.emitEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECTED))

    def onDisconnected(self):
        if self.state != self.__class__.STATE_DISCONNECTED:
            self.state = self.__class__.STATE_DISCONNECTED
            self.connected = False
            logger.debug("Disconnected")
            self.emitEvent(YowLayerEvent(self.__class__.EVENT_STATE_DISCONNECTED, reason="", detached=True))

    def onConnecting(self):
        pass

    def onConnectionError(self, error):
        self.onDisconnected()

    @EventCallback(EVENT_STATE_CONNECT)
    def onConnectLayerEvent(self, ev):
        if not self.connected:
            self.createConnection()
        else:
            logger.warn("Received connect event while already connected")
        return True

    @EventCallback(EVENT_STATE_DISCONNECT)
    def onDisconnectLayerEvent(self, ev):
        self.destroyConnection(ev.getArg("reason"))
        return True

    def createConnection(self):
        self._dispatcher = self.__create_dispatcher(self.getProp(self.PROP_DISPATCHER, self.DISPATCHER_DEFAULT))
        self.state = self.__class__.STATE_CONNECTING
        endpoint = self.getProp(self.__class__.PROP_ENDPOINT)
        logger.info("Connecting to %s:%s" % endpoint)
        self._dispatcher.connect(endpoint)

    def destroyConnection(self, reason = None):
        self.state = self.__class__.STATE_DISCONNECTING
        self._dispatcher.disconnect()

    def getStatus(self):
        return self.connected

    def send(self, data):
        if self.connected:
            self._dispatcher.sendData(data)

    def onRecvData(self, data):
        self.receive(data)

    def receive(self, data):
        self.toUpper(data)

    def __str__(self):
        return "Network Layer"



from axolotl.state.identitykeystore import IdentityKeyStore
from axolotl.identitykey import IdentityKey
from axolotl.identitykeypair import IdentityKeyPair
from axolotl.util.keyhelper import KeyHelper
from axolotl.ecc.djbec import *
import sys


class LiteIdentityKeyStore(IdentityKeyStore):
    def __init__(self, dbConn):
        """
        :type dbConn: Connection
        """
        self.dbConn = dbConn
        dbConn.execute("CREATE TABLE IF NOT EXISTS identities (_id INTEGER PRIMARY KEY AUTOINCREMENT,"
                       "recipient_id INTEGER UNIQUE,"
                       "registration_id INTEGER, public_key BLOB, private_key BLOB,"
                       "next_prekey_id INTEGER, timestamp INTEGER);")

        if self.getLocalRegistrationId() is None or self.getIdentityKeyPair() is None:
            identity = KeyHelper.generateIdentityKeyPair()
            registration_id = KeyHelper.generateRegistrationId(True)
            self._storeLocalData(registration_id, identity)

    def getIdentityKeyPair(self):
        q = "SELECT public_key, private_key FROM identities WHERE recipient_id = -1"
        c = self.dbConn.cursor()
        c.execute(q)
        result = c.fetchone()

        if result:
            publicKey, privateKey = result
            return IdentityKeyPair(IdentityKey(DjbECPublicKey(publicKey[1:])), DjbECPrivateKey(privateKey))
        return None

    def getLocalRegistrationId(self):
        q = "SELECT registration_id FROM identities WHERE recipient_id = -1"
        c = self.dbConn.cursor()
        c.execute(q)
        result = c.fetchone()
        return result[0] if result else None


    def _storeLocalData(self, registrationId, identityKeyPair):
        q = "INSERT INTO identities(recipient_id, registration_id, public_key, private_key) VALUES(-1, ?, ?, ?)"
        c = self.dbConn.cursor()
        pubKey = identityKeyPair.getPublicKey().getPublicKey().serialize()
        privKey = identityKeyPair.getPrivateKey().serialize()

        if sys.version_info < (2,7):
            pubKey = buffer(pubKey)
            privKey = buffer(privKey)

        c.execute(q, (registrationId,
                      pubKey,
                      privKey))

        self.dbConn.commit()

    def saveIdentity(self, recipientId, identityKey):
        q = "DELETE FROM identities WHERE recipient_id=?"
        self.dbConn.cursor().execute(q, (recipientId,))
        self.dbConn.commit()


        q = "INSERT INTO identities (recipient_id, public_key) VALUES(?, ?)"
        c = self.dbConn.cursor()

        pubKey = identityKey.getPublicKey().serialize()
        c.execute(q, (recipientId, buffer(pubKey) if sys.version_info < (2,7) else pubKey))
        self.dbConn.commit()

    def isTrustedIdentity(self, recipientId, identityKey):
        q = "SELECT public_key from identities WHERE recipient_id = ?"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId,))
        result = c.fetchone()
        if not result:
            return True

        pubKey = identityKey.getPublicKey().serialize()

        if sys.version_info < (2, 7):
            pubKey = buffer(pubKey)

        return result[0] == pubKey


from yowsup.layers.protocol_media.protocolentities.message_media_downloadable_video \
    import VideoDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message
from .test_message_media import MediaMessageProtocolEntityTest


class VideoDownloadableMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(VideoDownloadableMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = VideoDownloadableMediaMessageProtocolEntity
        proto_node = self.node.getChild("proto")
        m = Message()
        media_message = Message.VideoMessage()
        media_message.url = "url"
        media_message.mimetype = "video/mp4"
        media_message.caption = "caption"
        media_message.file_sha256 = b"SHA256"
        media_message.file_length = 123
        media_message.height = 20
        media_message.width = 20
        media_message.media_key = b"MEDIA_KEY"
        media_message.jpeg_thumbnail = b"THUMBNAIL"
        m.video_message.MergeFrom(media_message)
        proto_node.setData(m.SerializeToString())



from yowsup.config.base import serialize
from yowsup.config.transforms.filter import FilterTransform
from yowsup.config.transforms.meta import MetaPropsTransform
from yowsup.config.transforms.map import MapTransform
from yowsup.config.transforms.config_dict import ConfigDictTransform
from yowsup.config.transforms.props import PropsTransform

from consonance.structs.keypair import KeyPair
from consonance.structs.publickey import PublicKey
import base64


class ConfigSerialize(serialize.ConfigSerialize):
    def __init__(self, config_class):
        super(ConfigSerialize, self).__init__(
            transforms=(
                ConfigDictTransform(config_class),
                FilterTransform(
                    transform_filter=lambda key, val: val is not None,
                    reverse_filter=lambda key, val: key != "version"
                ),
                MapTransform(transform_map=lambda key, val: (key[1:], val)),
                PropsTransform(
                    transform_map={
                        "server_static_public": lambda key, val: (key, base64.b64encode(val.data).decode()),
                        "client_static_keypair": lambda key, val: (key, base64.b64encode(val.private.data + val.public.data).decode()),
                        "id": lambda key, val: (key, base64.b64encode(val).decode()),
                        "expid": lambda key, val: (key, base64.b64encode(val).decode()),
                        "edge_routing_info": lambda key, val: (key, base64.b64encode(val).decode())
                    },
                    reverse_map={
                        "server_static_public": lambda key, val: (key, PublicKey(base64.b64decode(val))),
                        "client_static_keypair": lambda key, val: (key, KeyPair.from_bytes(base64.b64decode(val))),
                        "id": lambda key, val: (key, base64.b64decode(val)),
                        "expid": lambda key, val: (key, base64.b64decode(val)),
                        "edge_routing_info": lambda key, val: (key, base64.b64decode(val))
                    }
                ),
                MetaPropsTransform(meta_props=("version", )),
            )
        )



from yowsup.layers import YowProtocolLayer
from yowsup.layers.auth.layer_authentication import YowAuthenticationProtocolLayer
from yowsup.layers.axolotl.protocolentities import *
from yowsup.axolotl.factory import AxolotlManagerFactory, AxolotlManager
from yowsup.layers.network.layer import YowNetworkLayer
from yowsup.layers import EventCallback

from yowsup.axolotl import exceptions
from yowsup.layers.axolotl.props import PROP_IDENTITY_AUTOTRUST

import logging
logger = logging.getLogger(__name__)


class AxolotlBaseLayer(YowProtocolLayer):
    def __init__(self):
        super(AxolotlBaseLayer, self).__init__()
        self._manager = None  # type: AxolotlManager | None
        self.skipEncJids = []

    def send(self, node):
        pass

    def receive(self, node):
        self.processIqRegistry(node)

    @property
    def manager(self):
        """
        :return:
        :rtype: AxolotlManager
        """
        return self._manager

    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def on_connected(self, yowLayerEvent):
        self._manager = AxolotlManagerFactory().get_manager(
                self.getProp(YowAuthenticationProtocolLayer.PROP_CREDENTIALS)[0]
        )

    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECTED)
    def on_disconnected(self, yowLayerEvent):
        self._manager = None

    def getKeysFor(self, jids, resultClbk, errorClbk = None, reason=None):
        logger.debug("getKeysFor(jids=%s, resultClbk=[omitted], errorClbk=[omitted], reason=%s)" % (jids, reason))

        def onSuccess(resultNode, getKeysEntity):
            entity = ResultGetKeysIqProtocolEntity.fromProtocolTreeNode(resultNode)
            resultJids = entity.getJids()
            successJids = []
            errorJids = {} #jid -> exception

            for jid in getKeysEntity.jids:
                if jid not in resultJids:
                    self.skipEncJids.append(jid)
                    continue

                recipient_id = jid.split('@')[0]
                preKeyBundle = entity.getPreKeyBundleFor(jid)
                try:
                    self.manager.create_session(recipient_id, preKeyBundle,
                                                autotrust=self.getProp(PROP_IDENTITY_AUTOTRUST, False))
                    successJids.append(jid)
                except exceptions.UntrustedIdentityException as e:
                        errorJids[jid] = e
                        logger.error(e)
                        logger.warning("Ignoring message with untrusted identity")

            resultClbk(successJids, errorJids)

        def onError(errorNode, getKeysEntity):
            if errorClbk:
                errorClbk(errorNode, getKeysEntity)

        entity = GetKeysIqProtocolEntity(jids, reason=reason)
        self._sendIq(entity, onSuccess, onError=onError)



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode


class GetKeysIqProtocolEntity(IqProtocolEntity):
    def __init__(self, jids, reason=None):
        super(GetKeysIqProtocolEntity, self).__init__("encrypt", _type="get", to=YowConstants.WHATSAPP_SERVER)
        self.jids = jids
        self.reason = reason

    @property
    def reason(self):
        # type: () -> str
        return self._reason

    @reason.setter
    def reason(self, value):
        # type: (str) -> None
        self._reason = value

    @property
    def jids(self):
        # type: () -> list[str]
        return self._jids

    @jids.setter
    def jids(self, value):
        # type: (list[str]) -> None
        assert type(value) is list, "expected list of jids, got %s" % type(value)
        self._jids = value

    def toProtocolTreeNode(self):
        node = super(GetKeysIqProtocolEntity, self).toProtocolTreeNode()
        keyNode = ProtocolTreeNode("key")

        for jid in self.jids:
            attrs = { "jid": jid }
            if self.reason is not None:
                attrs["reason"] = self.reason
            userNode = ProtocolTreeNode("user", attrs)
            keyNode.addChild(userNode)
        node.addChild(keyNode)
        return node



from yowsup.axolotl.manager import AxolotlManager
from yowsup.common.tools import StorageTools
from yowsup.axolotl.store.sqlite.liteaxolotlstore import LiteAxolotlStore
import logging

logger = logging.getLogger(__name__)


class AxolotlManagerFactory(object):
    DB = "axolotl.db"

    def get_manager(self, username):
        logger.debug("get_manager(username=%s)" % username)
        dbpath = StorageTools.constructPath(username, self.DB)
        store = LiteAxolotlStore(dbpath)
        return AxolotlManager(store, username)



from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
import socket
import ssl
import os
import hashlib
import sys
from time import sleep
import threading
import logging
from yowsup.common.tools import MimeTools

logger = logging.getLogger(__name__)


class MediaUploader(WARequest, threading.Thread):
    def __init__(self, jid, accountJid, sourcePath, uploadUrl, resumeOffset=0, successClbk=None, errorClbk=None, progressCallback=None, asynchronous=True):
        WARequest.__init__(self)

        self.asynchronous = asynchronous
        self.jid = jid
        self.accountJid = accountJid
        self.sourcePath = sourcePath
        self.uploadUrl = uploadUrl
        self.resumeOffset = resumeOffset

        self.successCallback = successClbk
        self.errorCallback = errorClbk
        self.progressCallback = progressCallback

        self.pvars = ["name", "type", "size", "url", "error",
                      "mimetype", "filehash", "width", "height"]

        self.setParser(JSONResponseParser())

        self.sock = socket.socket()

    def start(self):
        if self.asynchronous:
            threading.Thread.__init__(self)
            super(MediaUploader, self).start()
        else:
            self.run()

    def run(self):

        sourcePath = self.sourcePath
        uploadUrl = self.uploadUrl
        _host = uploadUrl.replace("https://", "")

        self.url = _host[:_host.index('/')]

        try:
            filename = os.path.basename(sourcePath)
            filetype = MimeTools.getMIME(filename)
            filesize = os.path.getsize(sourcePath)

            self.sock.connect((self.url, self.port))
            ssl_sock = ssl.wrap_socket(self.sock)

            m = hashlib.md5()
            m.update(filename.encode())
            crypto = m.hexdigest() + os.path.splitext(filename)[1]

            boundary = "zzXXzzYYzzXXzzQQ"  # "-------" + m.hexdigest() #"zzXXzzYYzzXXzzQQ"
            contentLength = 0

            hBAOS = "--" + boundary + "\r\n"
            hBAOS += "Content-Disposition: form-data; name=\"to\"\r\n\r\n"
            hBAOS += self.jid + "\r\n"
            hBAOS += "--" + boundary + "\r\n"
            hBAOS += "Content-Disposition: form-data; name=\"from\"\r\n\r\n"
            hBAOS += self.accountJid.replace("@whatsapp.net", "") + "\r\n"

            hBAOS += "--" + boundary + "\r\n"
            hBAOS += "Content-Disposition: form-data; name=\"file\"; filename=\"" + crypto + "\"\r\n"
            hBAOS += "Content-Type: " + filetype + "\r\n\r\n"

            fBAOS = "\r\n--" + boundary + "--\r\n"

            contentLength += len(hBAOS)
            contentLength += len(fBAOS)
            contentLength += filesize

            POST = "POST %s\r\n" % uploadUrl
            POST += "Content-Type: multipart/form-data; boundary=" + boundary + "\r\n"
            POST += "Host: %s\r\n" % self.url
            POST += "User-Agent: %s\r\n" % self.getUserAgent()
            POST += "Content-Length: " + str(contentLength) + "\r\n\r\n"

            ssl_sock.write(bytearray(POST.encode()))
            ssl_sock.write(bytearray(hBAOS.encode()))

            totalsent = 0
            buf = 1024
            f = open(sourcePath, 'rb')
            stream = f.read()
            f.close()
            status = 0
            lastEmit = 0

            while totalsent < int(filesize):
                ssl_sock.write(stream[:buf])
                status = totalsent * 100 / filesize
                if lastEmit != status and status != 100 and filesize > 12288:
                    if self.progressCallback:
                        self.progressCallback(self.sourcePath, self.jid, uploadUrl, int(status))
                lastEmit = status
                stream = stream[buf:]
                totalsent = totalsent + buf

            ssl_sock.write(bytearray(fBAOS.encode()))

            sleep(1)
            data = ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)
            data += ssl_sock.recv(8192)

            if self.progressCallback:
                self.progressCallback(self.sourcePath, self.jid, uploadUrl, 100)

            lines = data.decode().splitlines()

            result = None

            for l in lines:
                if l.startswith("{"):
                    result = self.parser.parse(l, self.pvars)
                    break

            if not result:
                raise Exception("json data not found")

            if result["url"] is not None:
                if self.successCallback:
                    self.successCallback(sourcePath, self.jid, result["url"])
            else:
                logger.exception("uploadUrl: %s, result of uploading media has no url" % uploadUrl)
                if self.errorCallback:
                    self.errorCallback(sourcePath, self.jid, uploadUrl)

        except:
            logger.exception("Error occured at transfer %s" % sys.exc_info()[1])
            if self.errorCallback:
                self.errorCallback(sourcePath, self.jid, uploadUrl)



from yowsup.config.base.transform import ConfigTransform
from yowsup.config.transforms.props import PropsTransform


class MetaPropsTransform(ConfigTransform):
    META_FORMAT = "__%s__"

    def __init__(self, meta_props=None, meta_format=META_FORMAT):
        meta_props = meta_props or ()
        meta_format = meta_format
        transform_map = {}
        reverse_map = {}
        for prop in meta_props:
            formatted = meta_format % prop
            transform_map[prop] = lambda key, val, formatted=formatted: (formatted, val)
            reverse_map[formatted] = lambda key, val, prop=prop: (prop, val)

        self._props_transform = PropsTransform(transform_map=transform_map, reverse_map=reverse_map)

    def transform(self, data):
        return self._props_transform.transform(data)

    def reverse(self, data):
        return self._props_transform.reverse(data)



from yowsup.layers.protocol_media.protocolentities.test_message_media import MediaMessageProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import LocationMediaMessageProtocolEntity
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message


class LocationMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(LocationMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = LocationMediaMessageProtocolEntity

        m = Message()
        location_message = Message.LocationMessage()
        location_message.degrees_latitude = 30.089037
        location_message.degrees_longitude = 31.319488
        location_message.name = "kaos"
        location_message.url = "kaos_url"

        m.location_message.MergeFrom(location_message)

        proto_node = self.node.getChild("proto")
        proto_node["mediatype"] = "location"
        proto_node.setData(m.SerializeToString())


from .layer import SyncLayer

from yowsup.stacks import  YowStackBuilder
from yowsup.layers import YowLayerEvent
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers.network import YowNetworkLayer


class YowsupSyncStack(object):
    def __init__(self, credentials, contacts):
        """
        :param credentials:
        :param contacts: list of [jid ]
        :return:
        """
        stackBuilder = YowStackBuilder()

        self._stack = stackBuilder \
            .pushDefaultLayers() \
            .push(SyncLayer) \
            .build()

        self._stack.setProp(SyncLayer.PROP_CONTACTS, contacts)
        self._stack.setProp(YowAuthenticationProtocolLayer.PROP_PASSIVE, True)
        self._stack.setCredentials(credentials)

    def set_prop(self, key, val):
        self._stack.setProp(key, val)

    def start(self):
        self._stack.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))
        self._stack.loop()



from consonance.protocol import WANoiseProtocol
from consonance.streams.segmented.segmented import SegmentedStream
from consonance.config.client import ClientConfig
from dissononce.dh.keypair import KeyPair
from dissononce.dh.x25519.x25519 import PublicKey

import threading


class WANoiseProtocolHandshakeWorker(threading.Thread):
    def __init__(self, wanoiseprotocol, stream, client_config, s, rs=None, finish_callback=None):
        """
        :param wanoiseprotocol:
        :type wanoiseprotocol: WANoiseProtocol
        :param stream:
        :type stream: SegmentedStream
        :param client_config:
        :type client_config: ClientConfig
        :param s:
        :type s: KeyPair
        :param rs:
        :type rs: PublicKey | None
        """
        super(WANoiseProtocolHandshakeWorker, self).__init__()
        self.daemon = True

        self._protocol = wanoiseprotocol # type: WANoiseProtocol
        self._stream = stream # type: SegmentedStream
        self._client_config = client_config # type: ClientConfig
        self._s = s # type: KeyPair
        self._rs = rs # type: PublicKey
        self._finish_callback = finish_callback

    def run(self):
        self._protocol.reset()
        self._protocol.start(self._stream, self._client_config, self._s, self._rs)
        if self._finish_callback is not None:
            self._finish_callback()



from axolotl.groups.state.senderkeystore import SenderKeyStore
from axolotl.groups.state.senderkeyrecord import SenderKeyRecord
import sqlite3
import sys
class LiteSenderKeyStore(SenderKeyStore):
    def __init__(self, dbConn):
        """
        :type dbConn: Connection
        """
        self.dbConn = dbConn
        dbConn.execute("CREATE TABLE IF NOT EXISTS sender_keys (_id INTEGER PRIMARY KEY AUTOINCREMENT,"
                       "group_id TEXT NOT NULL,"
                       "sender_id INTEGER NOT NULL, record BLOB);")

        dbConn.execute("CREATE UNIQUE INDEX IF NOT EXISTS sender_keys_idx ON sender_keys (group_id, sender_id);")

    def storeSenderKey(self, senderKeyName, senderKeyRecord):
        """
        :type senderKeyName: SenderKeName
        :type senderKeyRecord: SenderKeyRecord
        """
        q = "INSERT INTO sender_keys (group_id, sender_id, record) VALUES(?,?, ?)"
        cursor = self.dbConn.cursor()
        serialized = senderKeyRecord.serialize()
        if sys.version_info < (2,7):
            serialized = buffer(serialized)
        try:
            cursor.execute(q, (senderKeyName.getGroupId(), senderKeyName.getSender().getName(), serialized))
            self.dbConn.commit()
        except sqlite3.IntegrityError as e:
            q = "UPDATE sender_keys set record = ? WHERE group_id = ? and sender_id = ?"
            cursor = self.dbConn.cursor()
            cursor.execute(q, (serialized, senderKeyName.getGroupId(), senderKeyName.getSender().getName()))
            self.dbConn.commit()

    def loadSenderKey(self, senderKeyName):
        """
        :type senderKeyName: SenderKeyName
        """
        q = "SELECT record FROM sender_keys WHERE group_id = ? and sender_id = ?"
        cursor = self.dbConn.cursor()
        cursor.execute(q, (senderKeyName.getGroupId(), senderKeyName.getSender().getName()))

        result = cursor.fetchone()
        if not result:
            return SenderKeyRecord()
        return SenderKeyRecord(serialized = result[0])



from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes
from .message_media import MediaMessageProtocolEntity
from .attributes.attributes_contact import ContactAttributes
from .attributes.attributes_media import MediaAttributes


class ContactMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    def __init__(self, contact_attrs, media_attrs, message_attrs):
        # type: (ContactAttributes, MediaAttributes, MessageAttributes) -> None
        super(ContactMediaMessageProtocolEntity, self).__init__("contact", media_attrs, message_attrs)
        self.display_name = contact_attrs.display_name
        self.vcard = contact_attrs.vcard

    @property
    def proto(self):
        return self._proto.contact_message

    @property
    def display_name(self):
        return self.proto.display_name

    @display_name.setter
    def display_name(self, value):
        self.proto.display_name = value

    @property
    def vcard(self):
        return self.proto.vcard

    @vcard.setter
    def vcard(self, value):
        self.proto.vcard = value



from yowsup.layers.network.dispatcher.dispatcher import YowConnectionDispatcher
import socket
import logging

logger = logging.getLogger(__name__)


class SocketConnectionDispatcher(YowConnectionDispatcher):
    def __init__(self, connectionCallbacks):
        super(SocketConnectionDispatcher, self).__init__(connectionCallbacks)
        self.socket = None

    def connect(self, host):
        if not self.socket:
            self.socket = socket.socket()
            self.connectAndLoop(host)
        else:
            logger.error("Already connected?")

    def disconnect(self):
        if self.socket:
            try:
                self.socket.shutdown(socket.SHUT_WR)
                self.socket.close()
            except socket.error as e:
                logger.error(e)
                self.socket = None
                self.connectionCallbacks.onDisconnected()
        else:
            logger.error("Not connected?")

    def connectAndLoop(self, host):
        socket = self.socket
        self.connectionCallbacks.onConnecting()
        try:
            socket.connect(host)
            self.connectionCallbacks.onConnected()
            while True:
                data = socket.recv(1024)
                if len(data):
                    self.connectionCallbacks.onRecvData(data)
                else:
                    break
            self.connectionCallbacks.onDisconnected()
        except Exception as e:
            logger.error(e)
            self.connectionCallbacks.onConnectionError(e)
        finally:
            self.socket = None
            socket.close()

    def sendData(self, data):
        try:
            self.socket.send(data)
        except socket.error as e:
            logger.error(e)
            self.disconnect()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protocol.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protocol.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x0eprotocol.proto\"R\n\nMessageKey\x12\x12\n\nremote_jid\x18\x01 \x01(\t\x12\x0f\n\x07\x66rom_me\x18\x02 \x01(\x08\x12\n\n\x02id\x18\x03 \x01(\t\x12\x13\n\x0bparticipant\x18\x04 \x01(\t')
)




_MESSAGEKEY = _descriptor.Descriptor(
  name='MessageKey',
  full_name='MessageKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='remote_jid', full_name='MessageKey.remote_jid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='from_me', full_name='MessageKey.from_me', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='MessageKey.id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='participant', full_name='MessageKey.participant', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18,
  serialized_end=100,
)

DESCRIPTOR.message_types_by_name['MessageKey'] = _MESSAGEKEY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MessageKey = _reflection.GeneratedProtocolMessageType('MessageKey', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGEKEY,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:MessageKey)
  ))
_sym_db.RegisterMessage(MessageKey)


# @@protoc_insertion_point(module_scope)



from .protomessage import ProtomessageProtocolEntity
from .message import MessageAttributes
class TextMessageProtocolEntity(ProtomessageProtocolEntity):
    def __init__(self, body, messageAttributes=None, to=None):
        #flexible attributes for temp backwards compat
        assert(bool(messageAttributes) ^ bool(to)), "Either set messageAttributes, or to, and not both"
        if to:
            messageAttributes = MessageAttributes(recipient=to)
        super(TextMessageProtocolEntity, self).__init__("text", messageAttributes)
        self.setBody(body)

    @property
    def conversation(self):
        return self.proto.conversation

    @conversation.setter
    def conversation(self, value):
        self.proto.conversation = value

    def getBody(self):
        #obsolete
        return self.conversation

    def setBody(self, body):
        #obsolete
        self.conversation = body



from yowsup.layers.protocol_media.protocolentities.message_media import MediaMessageProtocolEntity
from yowsup.layers.protocol_messages.protocolentities.test_message import MessageProtocolEntityTest
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message


class MediaMessageProtocolEntityTest(MessageProtocolEntityTest):
    def setUp(self):
        super(MediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = MediaMessageProtocolEntity
        proto_node = ProtocolTreeNode("proto", {"mediatype": "image"}, None, Message().SerializeToString())
        self.node.addChild(proto_node)



from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_downloadablemedia \
    import DownloadableMediaMessageAttributes
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_document \
    import DocumentAttributes
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes


class DocumentDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    def __init__(self, document_attrs, downloadablemedia_attrs, message_attrs):
        """
        :type document_attrs: DocumentAttributes
        :type downloadablemedia_attrs: DownloadableMediaMessageAttributes
        :type message_attrs: MessageAttributes
        """
        super(DocumentDownloadableMediaMessageProtocolEntity, self).__init__(
            "document", downloadablemedia_attrs, message_attrs
        )
        self.file_name = document_attrs.file_name
        self.file_length = document_attrs.file_length
        if document_attrs.title is not None:
            self.title = document_attrs.title
        if document_attrs.page_count is not None:
            self.page_count = document_attrs.page_count
        if document_attrs.jpeg_thumbnail is not None:
            self.jpeg_thumbnail = document_attrs.jpeg_thumbnail

    @property
    def proto(self):
        return self._proto.document_message

    @property
    def file_name(self):
        return self.proto.file_name

    @file_name.setter
    def file_name(self, value):
        self.proto.file_name = value

    @property
    def file_length(self):
        return self.proto.file_length

    @file_length.setter
    def file_length(self, value):
        self.proto.file_length = value

    @property
    def title(self):
        return self.proto.title

    @title.setter
    def title(self, value):
        self.proto.title = value

    @property
    def page_count(self):
        return self.proto.page_count

    @page_count.setter
    def page_count(self, value):
        self.proto.page_count = value

    @property
    def jpeg_thumbnail(self):
        return self.proto.image_message.jpeg_thumbnail

    @jpeg_thumbnail.setter
    def jpeg_thumbnail(self, value):
        self.proto.image_message.jpeg_thumbnail = value



from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_downloadablemedia \
    import DownloadableMediaMessageAttributes
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_video import VideoAttributes
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes


class VideoDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    def __init__(self, video_attrs, downloadablemedia_attrs, message_attrs):
        """
        :type video_attrs: VideoAttributes
        :type downloadablemedia_attrs: DownloadableMediaMessageAttributes
        :type message_attrs: MessageAttributes
        """
        super(VideoDownloadableMediaMessageProtocolEntity, self).__init__(
            "video", downloadablemedia_attrs, message_attrs
        )
        self.width = video_attrs.width
        self.height = video_attrs.height
        self.seconds = video_attrs.seconds
        if video_attrs.caption is not None:
            self.caption = video_attrs.caption
        self.gif_playback = video_attrs.gif_playback if video_attrs.gif_playback else False
        if video_attrs.jpeg_thumbnail is not None:
            self.jpeg_thumbnail = video_attrs.jpeg_thumbnail
        if video_attrs.gif_attribution is not None:
            self.gif_attribution = video_attrs.gif_attribution
        if video_attrs.streaming_sidecar is not None:
            self.streaming_sidecar = video_attrs.streaming_sidecar

    @property
    def proto(self):
        return self._proto.video_message

    @property
    def width(self):
        return self._proto.width

    @width.setter
    def width(self, value):
        self.proto.width = value

    @property
    def height(self):
        return self.proto.height

    @height.setter
    def height(self, value):
        self.proto.height = value

    @property
    def seconds(self):
        return self.proto.seconds

    @seconds.setter
    def seconds(self, value):
        self.proto.seconds = value

    @property
    def gif_playback(self):
        return self.proto.gif_playback

    @gif_playback.setter
    def gif_playback(self, value):
        self.proto.gif_playback = value

    @property
    def jpeg_thumbnail(self):
        return self._proto.jpeg_thumbnail

    @jpeg_thumbnail.setter
    def jpeg_thumbnail(self, value):
        self.proto.jpeg_thumbnail = value

    @property
    def caption(self):
        return self.proto.caption

    @caption.setter
    def caption(self, value):
        self.proto.caption = value

    @property
    def gif_attribution(self):
        return self.proto.gif_attribution

    @gif_attribution.setter
    def gif_attribution(self, value):
        self.proto.gif_attributions = value

    @property
    def streaming_sidecar(self):
        return self.proto.streaming_sidecar

    @streaming_sidecar.setter
    def streaming_sidecar(self, value):
        self.proto.streaming_sidecar = value



from yowsup.config.base.transform import ConfigTransform
import json


class DictJsonTransform(ConfigTransform):
    def transform(self, data):
        return json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))

    def reverse(self, data):
        return json.loads(data)




class ConfigTransform(object):
    def transform(self, config):
        """
        :param config:
        :type config: yowsup.config.base.config.Config
        :return: dict
        :rtype:
        """

    def reverse(self, data):
        """
        :param data:
        :type data:
        :return:
        :rtype: yowsup.config.base.config.Config
        """



from yowsup.layers import YowLayer
from .encoder import WriteEncoder
from .decoder import ReadDecoder
from .tokendictionary import TokenDictionary


class YowCoderLayer(YowLayer):

    def __init__(self):
        YowLayer.__init__(self)
        tokenDictionary = TokenDictionary()
        self.writer = WriteEncoder(tokenDictionary)
        self.reader = ReadDecoder(tokenDictionary)
    
    def send(self, data):
        self.write(self.writer.protocolTreeNodeToBytes(data))

    def receive(self, data):
        node = self.reader.getProtocolTreeNode(bytearray(data))
        if node:
            self.toUpper(node)

    def write(self, i):
        if(type(i) in(list, tuple)):
            self.toLower(bytearray(i))
        else:
            self.toLower(bytearray([i]))

    def __str__(self):
        return "Coder Layer"



from yowsup.layers import YowProtocolLayer
from .protocolentities import TextMessageProtocolEntity
class YowMessagesProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "message": (self.recvMessageStanza, self.sendMessageEntity)
        }
        super(YowMessagesProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Messages Layer"

    def sendMessageEntity(self, entity):
        if entity.getType() == "text":
            self.entityToLower(entity)

    ###recieved node handlers handlers
    def recvMessageStanza(self, node):
        protoNode = node.getChild("proto")

        if protoNode:
           if protoNode and protoNode["mediatype"] is None:
               self.toUpper(TextMessageProtocolEntity.fromProtocolTreeNode(node))




class LocationAttributes(object):
    def __init__(self,
                 degrees_latitude, degrees_longitude,
                 name=None, address=None, url=None,
                 duration=None, accuracy_in_meters=None, speed_in_mps=None, degrees_clockwise_from_magnetic_north=None,
                 axolotl_sender_key_distribution_message=None, jpeg_thumbnail=None):
        """
        :param degrees_latitude: Actual location, Place
        :param degrees_longitude:  Actual location, Place
        :param name: Place
        :param address: Place
        :param url: Place
        :param duration:
        :param accuracy_in_meters:
        :param speed_in_mps:
        :param degrees_clockwise_from_magnetic_north:
        :param axolotl_sender_key_distribution_message:
        :param jpeg_thumbnail: Actual location, Place
        """

        self._degrees_latitude = degrees_latitude
        self._degrees_longitude = degrees_longitude
        self._name = name
        self._address = address
        self._url = url
        self._duration = duration
        self._accuracy_in_meters = accuracy_in_meters
        self._speed_in_mps = speed_in_mps
        self._degrees_clockwise_from_magnetic_north = degrees_clockwise_from_magnetic_north
        self._axolotl_sender_key_distribution_message = axolotl_sender_key_distribution_message
        self._jpeg_thumbnail = jpeg_thumbnail

    @property
    def degrees_latitude(self):
        return self._degrees_latitude

    @property
    def degrees_longitude(self):
        return self._degrees_longitude

    @property
    def name(self):
        return self._name

    @property
    def address(self):
        return self._address

    @property
    def url(self):
        return self._url

    @property
    def duration(self):
        return self._duration

    @property
    def accuracy_in_meters(self):
        return self._accuracy_in_meters

    @property
    def speed_in_mps(self):
        return self._speed_in_mps

    @property
    def degrees_clockwise_from_magnetic_north(self):
        return self._degrees_clockwise_from_magnetic_north

    @property
    def axolotl_sender_key_distribution_message(self):
        return self._axolotl_sender_key_distribution_message

    @property
    def jpeg_thumbnail(self):
        return self._jpeg_thumbnail



from yowsup.layers import YowProtocolLayer
from .protocolentities import ImageDownloadableMediaMessageProtocolEntity
from .protocolentities import AudioDownloadableMediaMessageProtocolEntity
from .protocolentities import VideoDownloadableMediaMessageProtocolEntity
from .protocolentities import DocumentDownloadableMediaMessageProtocolEntity
from .protocolentities import LocationMediaMessageProtocolEntity
from .protocolentities import ContactMediaMessageProtocolEntity
from .protocolentities import ResultRequestUploadIqProtocolEntity
from .protocolentities import MediaMessageProtocolEntity
from .protocolentities import ExtendedTextMediaMessageProtocolEntity
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity, ErrorIqProtocolEntity
import logging

logger = logging.getLogger(__name__)


class YowMediaProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "message": (self.recvMessageStanza, self.sendMessageEntity),
            "iq": (self.recvIq, self.sendIq)
        }
        super(YowMediaProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Media Layer"

    def sendMessageEntity(self, entity):
        if entity.getType() == "media":
            self.entityToLower(entity)

    def recvMessageStanza(self, node):
        if node.getAttributeValue("type") == "media":
            mediaNode = node.getChild("proto")
            if mediaNode.getAttributeValue("mediatype") == "image":
                entity = ImageDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("mediatype") in ("audio", "ptt"):
                entity = AudioDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("mediatype") in ("video", "gif"):
                entity = VideoDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("mediatype") == "location":
                entity = LocationMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("mediatype") == "contact":
                entity = ContactMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("mediatype") == "document":
                entity = DocumentDownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            elif mediaNode.getAttributeValue("mediatype") == "url":
                entity = ExtendedTextMediaMessageProtocolEntity.fromProtocolTreeNode(node)
                self.toUpper(entity)
            else:
                logger.warn("Unsupported mediatype: %s, will send receipts" % mediaNode.getAttributeValue("mediatype"))
                self.toLower(MediaMessageProtocolEntity.fromProtocolTreeNode(node).ack(True).toProtocolTreeNode())

    def sendIq(self, entity):
        """
        :type entity: IqProtocolEntity
        """
        if entity.getType() == IqProtocolEntity.TYPE_SET and entity.getXmlns() == "w:m":
            #media upload!
            self._sendIq(entity, self.onRequestUploadSuccess, self.onRequestUploadError)

    def recvIq(self, node):
        """
        :type node: ProtocolTreeNode
        """

    def onRequestUploadSuccess(self, resultNode, requestUploadEntity):
        self.toUpper(ResultRequestUploadIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onRequestUploadError(self, errorNode, requestUploadEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))



from .layer_base import AxolotlBaseLayer
from yowsup.layers import YowLayerEvent, EventCallback
from yowsup.layers.network.layer import YowNetworkLayer
from yowsup.layers.axolotl.protocolentities import *
from yowsup.layers.auth.layer_authentication import YowAuthenticationProtocolLayer
from yowsup.layers.protocol_acks.protocolentities import OutgoingAckProtocolEntity
from axolotl.util.hexutil import HexUtil
from axolotl.ecc.curve import Curve
import logging
import binascii

logger = logging.getLogger(__name__)


class AxolotlControlLayer(AxolotlBaseLayer):
    def __init__(self):
        super(AxolotlControlLayer, self).__init__()
        self._unsent_prekeys = []
        self._reboot_connection = False

    def send(self, node):
        self.toLower(node)

    def receive(self, protocolTreeNode):
        """
        :type protocolTreeNode: ProtocolTreeNode
        """
        if not self.processIqRegistry(protocolTreeNode):
            if protocolTreeNode.tag == "notification" and protocolTreeNode["type"] == "encrypt":
                if protocolTreeNode.getChild("count") is not None:
                    return self.onRequestKeysEncryptNotification(protocolTreeNode)
                elif protocolTreeNode.getChild("identity") is not None:
                    return self.onIdentityChangeEncryptNotification(protocolTreeNode)

            self.toUpper(protocolTreeNode)

    def onIdentityChangeEncryptNotification(self, protocoltreenode):
        entity = IdentityChangeEncryptNotification.fromProtocolTreeNode(protocoltreenode)
        ack = OutgoingAckProtocolEntity(
            protocoltreenode["id"], "notification", protocoltreenode["type"], protocoltreenode["from"]
        )
        self.toLower(ack.toProtocolTreeNode())
        self.getKeysFor([entity.getFrom(True)], resultClbk=lambda _,__: None, reason="identity")

    def onRequestKeysEncryptNotification(self, protocolTreeNode):
        entity = RequestKeysEncryptNotification.fromProtocolTreeNode(protocolTreeNode)
        ack = OutgoingAckProtocolEntity(protocolTreeNode["id"], "notification", protocolTreeNode["type"], protocolTreeNode["from"])
        self.toLower(ack.toProtocolTreeNode())
        self.flush_keys(
            self.manager.generate_signed_prekey(),
            self.manager.level_prekeys(force=True)
        )

    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def on_connected(self, yowLayerEvent):
        super(AxolotlControlLayer, self).on_connected(yowLayerEvent)
        self.manager.level_prekeys()
        self._unsent_prekeys.extend(self.manager.load_unsent_prekeys())
        if len(self._unsent_prekeys):
            self.setProp(YowAuthenticationProtocolLayer.PROP_PASSIVE, True)

    @EventCallback(YowAuthenticationProtocolLayer.EVENT_AUTHED)
    def onAuthed(self, yowLayerEvent):
        if yowLayerEvent.getArg("passive") and len(self._unsent_prekeys):
            logger.debug("SHOULD FLUSH KEYS %d NOW!!" % len(self._unsent_prekeys))
            self.flush_keys(
                self.manager.load_latest_signed_prekey(generate=True),
                self._unsent_prekeys[:], reboot_connection=True
            )
            self._unsent_prekeys = []

    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECTED)
    def on_disconnected(self, yowLayerEvent):
        super(AxolotlControlLayer, self).on_disconnected(yowLayerEvent)
        logger.debug(("Disconnected, reboot_connect? = %s" % self._reboot_connection))
        if self._reboot_connection:
            self._reboot_connection = False
            #we requested this disconnect in this layer to switch off passive
            #no need to traverse it to upper layers?
            self.setProp(YowAuthenticationProtocolLayer.PROP_PASSIVE, False)
            self.getLayerInterface(YowNetworkLayer).connect()

    def flush_keys(self, signed_prekey, prekeys, reboot_connection=False):
        """
        sends prekeys
        :return:
        :rtype:
        """
        preKeysDict = {}
        for prekey in prekeys:
            keyPair = prekey.getKeyPair()
            preKeysDict[self.adjustId(prekey.getId())] = self.adjustArray(keyPair.getPublicKey().serialize()[1:])

        signedKeyTuple = (self.adjustId(signed_prekey.getId()),
                          self.adjustArray(signed_prekey.getKeyPair().getPublicKey().serialize()[1:]),
                          self.adjustArray(signed_prekey.getSignature()))

        setKeysIq = SetKeysIqProtocolEntity(
            self.adjustArray(
                self.manager.identity.getPublicKey().serialize()[1:]
            ),
            signedKeyTuple,
            preKeysDict,
            Curve.DJB_TYPE,
            self.adjustId(self.manager.registration_id)
        )

        onResult = lambda _, __: self.on_keys_flushed(prekeys, reboot_connection=reboot_connection)
        self._sendIq(setKeysIq, onResult, self.onSentKeysError)

    def on_keys_flushed(self, prekeys, reboot_connection):
        self.manager.set_prekeys_as_sent(prekeys)
        if reboot_connection:
            self._reboot_connection = True
            self.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT))

    def onSentKeysError(self, errorNode, keysEntity):
        raise Exception("Sent keys were not accepted")

    def adjustArray(self, arr):
        return HexUtil.decodeHex(binascii.hexlify(arr))

    def adjustId(self, _id):
        _id = format(_id, 'x')
        zfiller = len(_id) if len(_id) % 2 == 0 else len(_id) + 1
        _id = _id.zfill(zfiller if zfiller > 6 else 6)
        # if len(_id) % 2:
        #     _id = "0" + _id
        return binascii.unhexlify(_id)



class WriteEncoder:

    def __init__(self, tokenDictionary):
        self.tokenDictionary = tokenDictionary

    def protocolTreeNodeToBytes(self, node):
        outBytes = [0] # flags
        self.writeInternal(node, outBytes)

        return outBytes

    def writeInternal(self, node, data):

        x = 1 + \
        (0 if node.attributes is None else len(node.attributes) * 2) + \
        (0 if not node.hasChildren() else 1) + \
        (0 if node.data is None else 1)

        self.writeListStart(x, data)


        self.writeString(node.tag, data)
        self.writeAttributes(node.attributes, data);


        if node.data is not None:
            self.writeBytes(node.data, data)

        if node.hasChildren():
            self.writeListStart(len(node.children), data);
            for c in node.children:
                self.writeInternal(c, data);


    def writeAttributes(self, attributes, data):
        if attributes is not None:
            for key, value in attributes.items():
                self.writeString(key, data);
                self.writeString(value, data, True);


    def writeBytes(self, bytes_, data, packed = False):
        bytes__ = []
        for b in bytes_:
            if type(b) is int:
                bytes__.append(b)
            else:
                bytes__.append(ord(b))


        size = len(bytes__)
        toWrite = bytes__
        if size >= 0x100000:
            data.append(254)
            self.writeInt31(size, data)
        elif size >= 0x100:
            data.append(253)
            self.writeInt20(size, data)
        else:
            r = None
            if packed:
                if size < 128:
                    r = self.tryPackAndWriteHeader(255, bytes__, data)
                    if r is None:
                        r = self.tryPackAndWriteHeader(251, bytes__, data)

            if r is None:
                data.append(252)
                self.writeInt8(size, data)
            else:
                toWrite = r

        data.extend(toWrite)

    def writeInt8(self, v, data):
        data.append(v & 0xFF)


    def writeInt16(self, v, data):
        data.append((v & 0xFF00) >> 8);
        data.append((v & 0xFF) >> 0);

    def writeInt20(self, v, data):
        data.append((0xF0000 & v) >> 16)
        data.append((0xFF00 & v) >> 8)
        data.append((v & 0xFF) >> 0)

    def writeInt24(self, v, data):
        data.append((v & 0xFF0000) >> 16)
        data.append((v & 0xFF00) >> 8)
        data.append((v & 0xFF) >> 0)

    def writeInt31(self, v, data):
        data.append((0x7F000000 & v) >> 24)
        data.append((0xFF0000 & v) >> 16)
        data.append((0xFF00 & v) >> 8)
        data.append((v & 0xFF) >> 0)

    def writeListStart(self, i, data):
        if i == 0:
            data.append(0)
        elif i < 256:
            data.append(248)
            self.writeInt8(i, data)
        else:
            data.append(249)
            self.writeInt16(i, data)

    def writeToken(self, token, data):
        if token <= 255 and token >=0:
            data.append(token)
        else:
            raise ValueError("Invalid token: %s" % token)


    def writeString(self, tag, data, packed = False):
        tok = self.tokenDictionary.getIndex(tag)
        if tok:
            index, secondary = tok
            if secondary:
                self.writeToken(236, data)
            self.writeToken(index, data)
        else:
            at = '@'.encode() if type(tag) == bytes else '@'
            try:
                atIndex = tag.index(at)

                if atIndex < 1:
                    raise ValueError("atIndex < 1")
                else:
                    server = tag[atIndex+1:]
                    user = tag[0:atIndex]
                    self.writeJid(user, server, data)
            except ValueError:
                self.writeBytes(self.encodeString(tag), data, packed)

    def encodeString(self, string):
        res = []

        if type(string) == bytes:
            for char in string:
                res.append(char)
        else:
            for char in string:
                res.append(ord(char))
        return res;

    def writeJid(self, user, server, data):
        data.append(250)
        if user is not None:
            self.writeString(user, data, True)
        else:
            self.writeToken(0, data)
        self.writeString(server, data)


    def tryPackAndWriteHeader(self, v, headerData, data):
        size = len(headerData)
        if size >= 128:
            return None

        arr = [0] * int((size + 1) / 2)
        for i in range(0, size):
            packByte = self.packByte(v, headerData[i])
            if packByte == -1:
                arr = []
                break
            n2 = int(i / 2)
            arr[n2] |= (packByte << 4 * (1 - i % 2))
        if len(arr) > 0:
            if size % 2 == 1:
                arr[-1] |= 15 #0xF
            data.append(v)
            self.writeInt8(size %2 << 7 | len(arr), data)
            return arr

        return None



    def packByte(self, v, n2):
        if v == 251:
            return self.packHex(n2)
        if v == 255:
            return self.packNibble(n2)
        return -1

    def packHex(self, n):
        if n in range(48, 58):
            return n - 48
        if n in range(65, 71):
            return 10 + (n - 65)
        return -1

    def packNibble(self, n):
        if n in (45, 46):
            return 10 + (n - 45)

        if n in range(48, 58):
            return n - 48

        return -1



class AudioAttributes(object):
    def __init__(self, seconds, ptt):
        """
        :param seconds: duration of audio playback in seconds
        :type seconds: int
        :param ptt: indicates whether this is a push-to-talk audio message
        :type ptt: bool
        """
        self._seconds = seconds  # type: int
        self._ptt = ptt  # type: bool

    @property
    def seconds(self):
        return self._seconds

    @property
    def ptt(self):
        return self._ptt



from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_receipts.protocolentities import IncomingReceiptProtocolEntity
from yowsup.layers.axolotl.protocolentities.iq_keys_get_result import ResultGetKeysIqProtocolEntity
class RetryIncomingReceiptProtocolEntity(IncomingReceiptProtocolEntity):

    '''
    <receipt type="retry" from="xxxxxxxxxxx@s.whatsapp.net" participant="" id="1415389947-12" t="1432833777">
        <retry count="1" t="1432833266" id="1415389947-12" v="1">
        </retry>
        <registration>
            HEX:xxxxxxxxx
        </registration>
    </receipt>
    '''

    def __init__(self, _id, jid, remoteRegistrationId, receiptTimestamp, retryTimestamp, v = 1, count = 1, participant = None, offline = None):
        super(RetryIncomingReceiptProtocolEntity, self).__init__(_id, jid, receiptTimestamp, offline=offline, type="retry", participant=participant)
        self.setRetryData(remoteRegistrationId, v,count, retryTimestamp)

    def setRetryData(self, remoteRegistrationId, v, count, retryTimestamp):
        self.remoteRegistrationId =  remoteRegistrationId
        self.v = int(v)
        self.count = int(count)
        self.retryTimestamp = int(retryTimestamp)

    def toProtocolTreeNode(self):
        node = super(RetryIncomingReceiptProtocolEntity, self).toProtocolTreeNode()

        retry = ProtocolTreeNode("retry", {
            "count": str(self.count),
            "id": self.getId(),
            "v": str(self.v),
            "t": str(self.retryTimestamp)
        })
        node.addChild(retry)
        registration = ProtocolTreeNode("registration", data=ResultGetKeysIqProtocolEntity._intToBytes(self.remoteRegistrationId))
        node.addChild(registration)
        return node

    def getRetryCount(self):
        return self.count

    def getRetryJid(self):
        return self.getParticipant() or self.getFrom()

    def __str__(self):
        out = super(RetryIncomingReceiptProtocolEntity, self).__str__()
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IncomingReceiptProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = RetryIncomingReceiptProtocolEntity
        retryNode = node.getChild("retry")
        entity.setRetryData(ResultGetKeysIqProtocolEntity._bytesToInt(node.getChild("registration").data), retryNode["v"], retryNode["count"], retryNode["t"])

        return entity



from yowsup.layers.protocol_media.protocolentities.test_message_media import MediaMessageProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import ExtendedTextMediaMessageProtocolEntity
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message


class ExtendedTextMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(ExtendedTextMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = ExtendedTextMediaMessageProtocolEntity

        m = Message()
        media_message = Message.ExtendedTextMessage()
        media_message.canonical_url = "url"
        media_message.text = "text"
        media_message.matched_text = "matched_text"
        media_message.description = "desc"
        media_message.title = "title"
        media_message.jpeg_thumbnail = b"thumb"
        m.extended_text_message.MergeFrom(media_message)

        proto_node = self.node.getChild("proto")
        proto_node["mediatype"] = "url"
        proto_node.setData(m.SerializeToString())

        self._entity = ExtendedTextMediaMessageProtocolEntity\
            .fromProtocolTreeNode(self.node)  # type: ExtendedTextMediaMessageProtocolEntity

    def test_properties(self):
        self.assertEqual("url", self._entity.canonical_url)
        self.assertEqual("text", self._entity.text)
        self.assertEqual("matched_text", self._entity.matched_text)
        self.assertEqual("desc", self._entity.description)
        self.assertEqual("title", self._entity.title)
        self.assertEqual(b"thumb", self._entity.jpeg_thumbnail)



class ContactAttributes(object):
    def __init__(self, display_name, vcard):
        self._display_name = display_name
        self._vcard = vcard

    @property
    def display_name(self):
        return self._display_name

    @property
    def vcard(self):
        return self._vcard
